"""

visualize_video_output.py

Render a folder of videos with bounding boxes to a new folder, based on a
detector output file.

"""

#%% Imports

import argparse
import os
import random
import cv2

from tqdm import tqdm
from PIL import Image
import numpy as np

from megadetector.data_management.annotations.annotation_constants import detector_bbox_category_id_to_name
from megadetector.detection.video_utils import run_callback_on_frames, default_fourcc, is_video_file
from megadetector.utils.path_utils import path_is_abs
from megadetector.utils.wi_taxonomy_utils import load_md_or_speciesnet_file
from megadetector.visualization.visualization_utils import render_detection_bounding_boxes


#%% Constants

# This will only be used if a category mapping is not available in the results file
DEFAULT_DETECTOR_LABEL_MAP = {
    str(k): v for k, v in detector_bbox_category_id_to_name.items()
}

DEFAULT_CLASSIFICATION_THRESHOLD = 0.4
DEFAULT_DETECTION_THRESHOLD = 0.15


#%% Classes

class VideoVisualizationOptions:
    """
    Options controlling the behavior of visualize_video_output()
    """

    def __init__(self):

        #: Confidence threshold for including detections
        self.confidence_threshold = DEFAULT_DETECTION_THRESHOLD

        #: Sample N videos to process (-1 for all videos)
        self.sample = -1

        #: Random seed for sampling
        self.random_seed = None

        #: Confidence threshold for including classifications
        self.classification_confidence_threshold = DEFAULT_CLASSIFICATION_THRESHOLD

        #: Frame rate for output videos. Either a float (fps) or 'auto' to calculate
        #: based on detection frame intervals
        self.rendering_fs = 'auto'

        #: Fourcc codec specification for video encoding
        self.fourcc = default_fourcc

        #: Skip frames before first and after last above-threshold detection
        self.trim_to_detections = False

# ...class VideoVisualizationOptions


#%% Support functions

def _get_video_output_framerate(video_entry, original_framerate, rendering_fs='auto'):
    """
    Calculate the appropriate output frame rate for a video based on detection frame numbers.

    Args:
        video_entry (dict): video entry from results file containing detections
        original_framerate (float): original frame rate of the video
        rendering_fs (str or float): 'auto' for automatic calculation, negative float for
            speedup factor, positive float for explicit fps

    Returns:
        float: calculated output frame rate
    """

    if rendering_fs != 'auto':

        if float(rendering_fs) < 0:

            # Negative value means speedup factor
            speedup_factor = abs(float(rendering_fs))
            if ('detections' not in video_entry) or (len(video_entry['detections']) == 0):
                # This is a bit arbitrary, but a reasonable thing to do when we have no basis
                # to determine the output frame rate
                return original_framerate * speedup_factor

            frame_numbers = []
            for detection in video_entry['detections']:
                if 'frame_number' in detection:
                    frame_numbers.append(detection['frame_number'])

            if len(frame_numbers) < 2:
                # This is a bit arbitrary, but a reasonable thing to do when we have no basis
                # to determine the output frame rate
                return original_framerate * speedup_factor

            frame_numbers = sorted(set(frame_numbers))
            first_interval = frame_numbers[1] - frame_numbers[0]

            # Calculate base output frame rate based on first interval, then apply speedup
            base_output_fps = original_framerate / first_interval
            return base_output_fps * speedup_factor

        else:

            # Positive value means explicit fps
            return float(rendering_fs)

        # ...if we're using an explicit/speedup-based frame rate

    # ...if we aren't in "auto" frame rate mode

    # Auto mode
    if 'detections' not in video_entry or len(video_entry['detections']) == 0:
        return original_framerate

    frame_numbers = []
    for detection in video_entry['detections']:
        if 'frame_number' in detection:
            frame_numbers.append(detection['frame_number'])

    if len(frame_numbers) < 2:
        return original_framerate

    frame_numbers = sorted(set(frame_numbers))
    first_interval = frame_numbers[1] - frame_numbers[0]

    # Calculate output frame rate based on first interval
    output_fps = original_framerate / first_interval

    return output_fps


def _get_frames_to_process(video_entry, confidence_threshold, trim_to_detections=False):
    """
    Get list of frame numbers that have detections for this video.

    Args:
        video_entry (dict): video entry from results file
        confidence_threshold (float): minimum confidence for detections to be considered
        trim_to_detections (bool): if True, only include frames between first and last
            above-threshold detections (inclusive)

    Returns:
        list: sorted list of unique frame numbers to process
    """

    if 'detections' not in video_entry:
        return []

    if 'frames_processed' in video_entry:
        frame_numbers = set(video_entry['frames_processed'])
    else:
        frame_numbers = set()

    for detection in video_entry['detections']:

        if 'frame_number' in detection:
            # If this file includes the list of frames processed (required as of format
            # version 1.5), every frame with detections should be included in that list
            if 'frames_processed' in video_entry:
                if detection['frame_number'] not in frame_numbers:
                    print('Warning: frames_processed field present in {}, but frame {} is missing'.\
                          format(video_entry['file'],detection['frame_number']))
            frame_numbers.add(detection['frame_number'])
        else:
            print('Warning: detections in {} lack frame numbers'.format(video_entry['file']))

    # ...for each detection

    frame_numbers = sorted(list(frame_numbers))

    if trim_to_detections and (len(frame_numbers) > 0):

        # Find first and last frames with above-threshold detections

        above_threshold_frames = set()
        for detection in video_entry['detections']:
            if detection['conf'] >= confidence_threshold:
                above_threshold_frames.add(detection['frame_number'])

        if len(above_threshold_frames) > 0:

            above_threshold_frames = sorted(list(above_threshold_frames))
            first_detection_frame = above_threshold_frames[0]
            last_detection_frame = above_threshold_frames[-1]

            # Return all frames between first and last above-threshold detections (inclusive)
            trimmed_frames = []
            for frame_num in frame_numbers:
                if (first_detection_frame <= frame_num) and (frame_num <= last_detection_frame):
                    trimmed_frames.append(frame_num)
            return trimmed_frames

        else:
            # No above-threshold detections, return empty list
            return []

    # ...if we're supposed to be trimming to non-empty frames

    return frame_numbers


def _get_detections_for_frame(video_entry, frame_number, confidence_threshold):
    """
    Get all detections for a specific frame that meet confidence thresholds.

    Args:
        video_entry (dict): video entry from results file
        frame_number (int): frame number to get detections for
        confidence_threshold (float): minimum detection confidence

    Returns:
        list: list of detection dictionaries for this frame
    """

    if 'detections' not in video_entry:
        return []

    frame_detections = []

    for detection in video_entry['detections']:
        if ((detection['frame_number'] == frame_number) and
            (detection['conf'] >= confidence_threshold)):
            frame_detections.append(detection)

    return frame_detections


def _process_video(video_entry,
                   detector_label_map,
                   classification_label_map,
                   options,
                   video_dir,
                   out_dir):
    """
    Process a single video, rendering detections on frames and creating output video.

    Args:
        video_entry (dict): video entry from results file
        detector_label_map (dict): mapping of detection category IDs to names
        classification_label_map (dict): mapping of classification category IDs to names
        options (VideoVisualizationOptions): processing options
        video_dir (str): input video directory
        out_dir (str): output directory

    Returns:
        dict: processing result information
    """

    result = {
        'file': video_entry['file'],
        'success': False,
        'error': None,
        'frames_processed': 0
    }

    # Handle failure cases
    if ('failure' in video_entry) and (video_entry['failure'] is not None):
        result['error'] = 'Ignoring failed video: {}'.format(video_entry['failure'])
        return result

    # Construct input and output paths
    if video_dir is None:
        input_video_path = video_entry['file']
        assert path_is_abs(input_video_path), \
            'Absolute paths are required when no video base dir is supplied'
    else:
        assert not path_is_abs(video_entry['file']), \
            'Relative paths are required when a video base dir is supplied'
        input_video_path = os.path.join(video_dir, video_entry['file'])

    if not os.path.exists(input_video_path):
        result['error'] = 'Video not found: {}'.format(input_video_path)
        return result

    # Create output path preserving directory structure
    rel_path = video_entry['file']
    output_video_path = os.path.join(out_dir, rel_path)
    os.makedirs(os.path.dirname(output_video_path), exist_ok=True)

    # Get frames to process
    frames_to_process = _get_frames_to_process(video_entry,
                                               options.confidence_threshold,
                                               options.trim_to_detections)
    if len(frames_to_process) == 0:
        result['error'] = 'No frames with detections to process'
        return result

    # Determine output frame rate
    original_framerate = video_entry['frame_rate']
    output_framerate = _get_video_output_framerate(video_entry,
                                                   original_framerate,
                                                   options.rendering_fs)

    # Storage for rendered frames
    rendered_frames = []

    def frame_callback(frame_array, frame_id):
        """
        Callback function for processing each frame.

        Args:
            frame_array (np.array): frame image data
            frame_id (str): frame identifier (unused)

        Returns:
            np.array: processed frame
        """

        # Extract frame number from the current processing context
        current_frame_idx = len(rendered_frames)
        if current_frame_idx >= len(frames_to_process):
            print('Warning: received an extra frame (index {} of {}) for video {}'.format(
                current_frame_idx,len(frames_to_process),video_entry['file']
            ))
            return frame_array

        current_frame_number = frames_to_process[current_frame_idx]

        # Convert numpy array to PIL Image
        if frame_array.dtype != np.uint8:
            frame_array = (frame_array * 255).astype(np.uint8)

        # Convert from BGR (OpenCV) to RGB (PIL) if needed
        if len(frame_array.shape) == 3 and frame_array.shape[2] == 3:
            frame_array = cv2.cvtColor(frame_array, cv2.COLOR_BGR2RGB)

        pil_image = Image.fromarray(frame_array)

        # Get detections for this frame
        frame_detections = _get_detections_for_frame(
            video_entry,
            current_frame_number,
            options.confidence_threshold
        )

        # Render detections on the frame
        if frame_detections:
            render_detection_bounding_boxes(
                frame_detections,
                pil_image,
                detector_label_map,
                classification_label_map,
                classification_confidence_threshold=options.classification_confidence_threshold
            )

        # Convert back to numpy array for video writing
        frame_array = np.array(pil_image)
        if (len(frame_array.shape) == 3) and (frame_array.shape[2] == 3):
            frame_array = cv2.cvtColor(frame_array, cv2.COLOR_RGB2BGR)

        rendered_frames.append(frame_array)
        return frame_array

    # ...def frame_callback(...)

    # Process video frames
    try:
        run_callback_on_frames(
            input_video_path,
            frame_callback,
            frames_to_process=frames_to_process,
            verbose=False
        )
    except Exception as e:
        import traceback
        trace = traceback.format_exc()
        result['error'] = 'Error processing video frames: {} ({})'.format(str(e),trace)
        return result

    # Write output video
    if len(rendered_frames) > 0:

        try:

            # Get frame dimensions
            height, width = rendered_frames[0].shape[:2]

            # Create VideoWriter
            fourcc = cv2.VideoWriter_fourcc(*options.fourcc)
            video_writer = cv2.VideoWriter(output_video_path, fourcc, output_framerate, (width, height))

            if not video_writer.isOpened():
                result['error'] = 'Failed to open video writer for {}'.format(output_video_path)
                return result

            # Write frames
            for frame in rendered_frames:
                video_writer.write(frame)

            video_writer.release()
            result['success'] = True
            result['frames_processed'] = len(rendered_frames)

        except Exception as e:

            result['error'] = 'Error writing output video: {}'.format(str(e))
            return result

        # ...try/except

    else:

        result['error'] = 'No frames were processed for video {}'.format(video_entry['file'])

    return result

# ...def _process_video(...)


#%% Main function

def visualize_video_output(detector_output_path,
                           out_dir,
                           video_dir,
                           options=None):
    """
    Renders videos with bounding boxes based on detector output.

    Args:
        detector_output_path (str): path to .json file containing detection results
        out_dir (str): output directory for rendered videos
        video_dir (str): input video directory
        options (VideoVisualizationOptions, optional): processing options

    Returns:
        list: list of processing results for each video
    """

    if options is None:
        options = VideoVisualizationOptions()

    # Validate that input and output directories are different
    if (video_dir is not None) and (os.path.abspath(out_dir) == os.path.abspath(video_dir)):
        raise ValueError('Output directory cannot be the same as video directory')

    # Load results file
    print('Loading results from {}'.format(detector_output_path))
    results_data = load_md_or_speciesnet_file(detector_output_path)

    # Get label mappings
    detector_label_map = results_data.get('detection_categories', DEFAULT_DETECTOR_LABEL_MAP)
    classification_label_map = results_data.get('classification_categories', {})

    # Filter to video entries only
    video_entries = []
    for entry in results_data['images']:
        if is_video_file(entry['file']):
            video_entries.append(entry)

    print('Found {} videos in results file'.format(len(video_entries)))

    # Apply sampling if requested
    if (options.sample > 0) and (len(video_entries) > options.sample):
        if options.random_seed is not None:
            random.seed(options.random_seed)
        n_videos_available = len(video_entries)
        video_entries = random.sample(video_entries, options.sample)
        print('Sampled {} of {} videos for processing'.format(
            len(video_entries),n_videos_available))

    # Create output directory
    os.makedirs(out_dir, exist_ok=True)

    # Process each video
    results = []

    for video_entry in tqdm(video_entries, desc='Processing videos'):
        result = _process_video(
            video_entry,
            detector_label_map,
            classification_label_map,
            options,
            video_dir,
            out_dir
        )
        results.append(result)

        if not result['success']:
            print('Warning: Failed to process {}: {}'.format(result['file'],result['error']))

    # ...for each video

    # Print summary
    successful = sum(1 for r in results if r['success'])
    failed = len(results) - successful
    total_frames = sum(r['frames_processed'] for r in results if r['success'])

    print('\nProcessing complete:')
    print(f'  Successfully processed: {successful} videos')
    print(f'  Failed: {failed} videos')
    print(f'  Total frames rendered: {total_frames}')

    return results

# ...def visualize_video_output(...)


#%% Command-line driver

def main():
    """
    Command-line driver for visualize_video_output
    """

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Render videos with bounding boxes predicted by a detector above '
                    'a confidence threshold, and save the rendered videos.')

    parser.add_argument(
        'detector_output_path',
        type=str,
        help='Path to json output file of the detector')

    parser.add_argument(
        'out_dir',
        type=str,
        help='Path to directory where the rendered videos will be saved. '
             'The directory will be created if it does not exist.')

    parser.add_argument(
        'video_dir',
        type=str,
        help='Path to directory containing the input videos')

    parser.add_argument(
        '--confidence_threshold',
        type=float,
        default=DEFAULT_DETECTION_THRESHOLD,
        help='Confidence threshold above which detections will be rendered')

    parser.add_argument(
        '--sample',
        type=int,
        default=-1,
        help='Number of videos to randomly sample for processing. '
             'Set to -1 to process all videos')

    parser.add_argument(
        '--random_seed',
        type=int,
        default=None,
        help='Random seed for reproducible sampling')

    parser.add_argument(
        '--classification_confidence_threshold',
        type=float,
        default=DEFAULT_CLASSIFICATION_THRESHOLD,
        help='Value between 0 and 1, indicating the confidence threshold '
             'above which classifications will be rendered')

    parser.add_argument(
        '--rendering_fs',
        default='auto',
        help='Frame rate for output videos. Use "auto" to calculate based on '
             'detection frame intervals, positive float for explicit fps, '
             'or negative float for speedup factor (e.g. -2.0 = 2x faster)')

    parser.add_argument(
        '--fourcc',
        type=str,
        default=default_fourcc,
        help='Fourcc codec specification for video encoding')

    parser.add_argument(
        '--trim_to_detections',
        action='store_true',
        help='Skip frames before first and after last above-threshold detection')

    args = parser.parse_args()

    # Create options object
    options = VideoVisualizationOptions()
    options.confidence_threshold = args.confidence_threshold
    options.sample = args.sample
    options.random_seed = args.random_seed
    options.classification_confidence_threshold = args.classification_confidence_threshold
    options.rendering_fs = args.rendering_fs
    options.fourcc = args.fourcc
    options.trim_to_detections = args.trim_to_detections

    # Run visualization
    visualize_video_output(
        args.detector_output_path,
        args.out_dir,
        args.video_dir,
        options
    )


if __name__ == '__main__':
    main()



"""

detect_and_crop.py

Run MegaDetector on images via Batch API, then save crops of the detected
bounding boxes.

The input to this script is a "queried images" JSON file, whose keys are paths
to images and values are dicts containing information relevant for training
a classifier, including labels and (optionally) ground-truth bounding boxes.
The image paths are in the format `<dataset-name>/<blob-name>` where we assume
that the dataset name does not contain '/'.

{
    "caltech/cct_images/59f79901-23d2-11e8-a6a3-ec086b02610b.jpg": {
        "dataset": "caltech",
        "location": 13,
        "class": "mountain_lion",  # class from dataset
        "bbox": [{"category": "animal",
                  "bbox": [0, 0.347, 0.237, 0.257]}],   # ground-truth bbox
        "label": ["monutain_lion"]  # labels to use in classifier
    },
    "caltech/cct_images/59f5fe2b-23d2-11e8-a6a3-ec086b02610b.jpg": {
        "dataset": "caltech",
        "location": 13,
        "class": "mountain_lion",  # class from dataset
        "label": ["monutain_lion"]  # labels to use in classifier
    },
    ...
}

We assume that no image contains over 100 bounding boxes, and we always save
crops as RGB .jpg files for consistency. For each image, each bounding box is
cropped and saved to a file with a suffix "___cropXX.jpg" (ground truth bbox) or
"___cropXX_mdvY.Y.jpg" (detected bbox) added to the filename of the original
image. "XX" ranges from "00" to "99" and "Y.Y" indicates the MegaDetector
version. If an image has ground truth bounding boxes, we assume that they are
exhaustive--i.e., there are no other objects of interest, so we don't need to
run MegaDetector on the image. If an image does not have ground truth bounding
boxes, we run MegaDetector on the image and label the detected boxes in order
from 00 up to 99. Based on the given confidence threshold, we may skip saving
certain bounding box crops, but we still increment the bounding box number for
skipped boxes.

Example cropped image path (with ground truth bbox from MegaDB)

    "path/to/crops/image.jpg___crop00.jpg"

Example cropped image path (with MegaDetector bbox)

    "path/to/crops/image.jpg___crop00_mdv4.1.jpg"

By default, the images are cropped exactly per the given bounding box
coordinates. However, if square crops are desired, pass the --square-crops
flag. This will always generate a square crop whose size is the larger of the
bounding box width or height. In the case that the square crop boundaries exceed
the original image size, the crop is padded with 0s.

This script currently only supports running MegaDetector via the Batch Detection
API. See the classification README for instructions on running MegaDetector
locally. If running the Batch Detection API, set the following environment
variables for the Azure Blob Storage container in which we save the intermediate
task lists:

    BATCH_DETECTION_API_URL                  # API URL
    CLASSIFICATION_BLOB_STORAGE_ACCOUNT      # storage account name
    CLASSIFICATION_BLOB_CONTAINER            # container name
    CLASSIFICATION_BLOB_CONTAINER_WRITE_SAS  # SAS token, without leading '?'
    DETECTION_API_CALLER                     # allow-listed API caller

This script allows specifying a directory where MegaDetector outputs are cached
via the --detector-output-cache-dir argument. This directory must be
organized as:

  <cache-dir>/<MegaDetector-version>/<dataset-name>.json

    Example: If the `cameratrapssc/classifier-training` Azure blob storage
    container is mounted to the local machine via blobfuse, it may be used as
    a MegaDetector output cache directory by passing
        "cameratrapssc/classifier-training/mdcache/"
    as the value for --detector-output-cache-dir.

This script outputs either 1 or 3 files, depending on whether the Batch Detection API
is run:

- <output_dir>/detect_and_crop_log_{timestamp}.json
    log of images missing detections and images that failed to properly
    download and crop
- <output_dir>/batchapi_tasklists/{task_id}.json
    (if --run-detector) task lists uploaded to the Batch Detection API
- <output_dir>/batchapi_response/{task_id}.json
    (if --run-detector) task status responses for completed tasks

"""

#%% Imports

from __future__ import annotations

import argparse
from collections.abc import Collection, Iterable, Mapping, Sequence
from concurrent import futures
from datetime import datetime
import json
import os
import pprint
import time
from typing import Any, Optional

from azure.storage.blob import ContainerClient
import requests
from tqdm import tqdm

from api.batch_processing.data_preparation.prepare_api_submission import (
    BatchAPIResponseError, Task, TaskStatus, divide_list_into_tasks)
from megadetector.classification.cache_batchapi_outputs import cache_detections
from megadetector.classification.crop_detections import load_and_crop
from megadetector.data_management.megadb import megadb_utils
from megadetector.utils import path_utils
from megadetector.utils import sas_blob_utils
from megadetector.utils import ct_utils


#%% Example usage

"""
    python detect_and_crop.py \
        base_logdir/queried_images.json \
        base_logdir \
        --detector-output-cache-dir /path/to/classifier-training/mdcache \
        --detector-version 4.1 \
        --run-detector --resume-file base_logdir/resume.json \
        --cropped-images-dir /path/to/crops --square-crops --threshold 0.9 \
        --save-full-images --images-dir /path/to/images --threads 50
"""


#%% Main function

def main(queried_images_json_path: str,
         output_dir: str,
         detector_version: str,
         detector_output_cache_base_dir: str,
         run_detector: bool,
         resume_file_path: Optional[str],
         cropped_images_dir: Optional[str],
         save_full_images: bool,
         square_crops: bool,
         check_crops_valid: bool,
         confidence_threshold: float,
         images_dir: Optional[str],
         threads: int) -> None:
    """
    Args:
        queried_images_json_path: str, path to output of json_validator.py
        detector_version: str, detector version string, e.g., '4.1',
            see {batch_detection_api_url}/supported_model_versions,
            determines the subfolder of detector_output_cache_base_dir in
            which to find and save detector outputs
        detector_output_cache_base_dir: str, path to local directory
            where detector outputs are cached, 1 JSON file per dataset
        cropped_images_dir: str, path to local directory for saving crops of
            bounding boxes
        run_detector: bool, whether to run Batch Detection API, or to skip
            running the detector entirely
        output_dir: str, path to directory to save outputs, see module docstring
        save_full_images: bool, whether to save downloaded images to images_dir,
            images_dir must be given if save_full_images=True
        square_crops: bool, whether to crop bounding boxes as squares
        check_crops_valid: bool, whether to load each crop to ensure the file is
            valid (i.e., not truncated)
        confidence_threshold: float, only crop bounding boxes above this value
        images_dir: optional str, path to local directory where images are saved
        threads: int, number of threads to use for downloading images
        resume_file_path: optional str, path to save JSON file with list of info
            dicts on running tasks, or to resume from running tasks, only used
            if run_detector=True
    """

    # This dictionary will get written out at the end of this process; store
    # diagnostic variables here
    log: dict[str, Any] = {}

    # error checking
    assert 0 <= confidence_threshold <= 1
    if save_full_images:
        assert images_dir is not None
        if not os.path.exists(images_dir):
            os.makedirs(images_dir, exist_ok=True)
            print(f'Created images_dir at {images_dir}')

    with open(queried_images_json_path, 'r') as f:
        js = json.load(f)
    detector_output_cache_dir = os.path.join(
        detector_output_cache_base_dir, f'v{detector_version}')
    if not os.path.exists(detector_output_cache_dir):
        os.makedirs(detector_output_cache_dir)
        print(f'Created directory at {detector_output_cache_dir}')
    images_without_ground_truth_bbox = [k for k in js if 'bbox' not in js[k]]
    images_to_detect, detection_cache, categories = filter_detected_images(
        potential_images_to_detect=images_without_ground_truth_bbox,
        detector_output_cache_dir=detector_output_cache_dir)
    print(f'{len(images_to_detect)} images not in detection cache')

    if run_detector:
        log['images_submitted_for_detection'] = images_to_detect

        assert resume_file_path is not None
        assert not os.path.isdir(resume_file_path)
        batch_detection_api_url = os.environ['BATCH_DETECTION_API_URL']

        if os.path.exists(resume_file_path):
            tasks_by_dataset = resume_tasks(
                resume_file_path,
                batch_detection_api_url=batch_detection_api_url)
        else:
            task_lists_dir = os.path.join(output_dir, 'batchapi_tasklists')
            tasks_by_dataset = submit_batch_detection_api(
                images_to_detect=images_to_detect,
                task_lists_dir=task_lists_dir,
                detector_version=detector_version,
                account=os.environ['CLASSIFICATION_BLOB_STORAGE_ACCOUNT'],
                container=os.environ['CLASSIFICATION_BLOB_CONTAINER'],
                sas_token=os.environ['CLASSIFICATION_BLOB_CONTAINER_WRITE_SAS'],
                caller=os.environ['DETECTION_API_CALLER'],
                batch_detection_api_url=batch_detection_api_url,
                resume_file_path=resume_file_path)

        wait_for_tasks(tasks_by_dataset, detector_output_cache_dir,
                       output_dir=output_dir)

        # refresh detection cache
        print('Refreshing detection cache...')
        images_to_detect, detection_cache, categories = filter_detected_images(
            potential_images_to_detect=images_without_ground_truth_bbox,
            detector_output_cache_dir=detector_output_cache_dir)
        print(f'{len(images_to_detect)} images not in detection cache')

    log['images_missing_detections'] = images_to_detect

    if cropped_images_dir is not None:

        images_failed_dload_crop, num_downloads, num_crops = download_and_crop(
            queried_images_json=js,
            detection_cache=detection_cache,
            detection_categories=categories,
            detector_version=detector_version,
            cropped_images_dir=cropped_images_dir,
            confidence_threshold=confidence_threshold,
            save_full_images=save_full_images,
            square_crops=square_crops,
            check_crops_valid=check_crops_valid,
            images_dir=images_dir,
            threads=threads,
            images_missing_detections=images_to_detect)
        log['images_failed_download_or_crop'] = images_failed_dload_crop
        log['num_new_downloads'] = num_downloads
        log['num_new_crops'] = num_crops

    print(f'{len(images_to_detect)} images with missing detections.')
    if cropped_images_dir is not None:
        print(f'{len(images_failed_dload_crop)} images failed to download or '
              'crop.')

    # save log of bad images
    date = datetime.now().strftime('%Y%m%d_%H%M%S')  # e.g., '20200722_110816'
    log_path = os.path.join(output_dir, f'detect_and_crop_log_{date}.json')
    ct_utils.write_json(log_path, log)


#%% Support functions

def load_detection_cache(detector_output_cache_dir: str,
                         datasets: Collection[str]) -> tuple[
                             dict[str, dict[str, dict[str, Any]]],
                             dict[str, str]
                         ]:
    """
    Loads detection cache for a given dataset. Returns empty dictionaries
    if the cache does not exist.

    Args:
        detector_output_cache_dir: str, path to local directory where detector
            outputs are cached, 1 JSON file per dataset
        datasets: list of str, names of datasets

    Returns:
        detection_cache: dict, maps dataset name to dict, which maps
            image file to corresponding entry in 'images' list from the
            Batch Detection API output. detection_cache[ds] is an empty dict
            if no cached detections were found for the given dataset ds.
        detection_categories: dict, maps str category ID to str category name
    """

    # cache of Detector outputs: dataset name => {img_path => detection_dict}
    detection_cache = {}
    detection_categories: dict[str, str] = {}

    pbar = tqdm(datasets)
    for ds in pbar:
        pbar.set_description(f'Loading dataset {ds} into detection cache')
        cache_path = os.path.join(detector_output_cache_dir, f'{ds}.json')
        if os.path.exists(cache_path):
            with open(cache_path, 'r') as f:
                js = json.load(f)
            detection_cache[ds] = {img['file']: img for img in js['images']}
            if len(detection_categories) == 0:
                detection_categories = js['detection_categories']
            assert detection_categories == js['detection_categories']
        else:
            tqdm.write(f'No detection cache found for dataset {ds}')
            detection_cache[ds] = {}
    return detection_cache, detection_categories


def filter_detected_images(
        potential_images_to_detect: Iterable[str],
        detector_output_cache_dir: str
        ) -> tuple[list[str],
                   dict[str, dict[str, dict[str, Any]]],
                   dict[str, str]]:
    """
    Checks image paths against cached Detector outputs, and prepares
    the SAS URIs for each image not in the cache.

    Args:
        potential_images_to_detect: list of str, paths to images that do not
            have ground truth bounding boxes, each path has format
            <dataset-name>/<img-filename>, where <img-filename> is the blob name
        detector_output_cache_dir: str, path to local directory where detector
            outputs are cached, 1 JSON file per dataset

    Returns:
        images_to_detect: list of str, paths to images not in the detector
            output cache, with the format <dataset-name>/<img-filename>
        detection_cache: dict, maps str dataset name to dict,
            detection_cache[dataset_name] is the 'detections' list from the
            Batch Detection API output
        detection_categories: dict, maps str category ID to str category name,
            empty dict if no cached detections are found
    """

    datasets = set(img_path[:img_path.find('/')]
                   for img_path in potential_images_to_detect)
    detection_cache, detection_categories = load_detection_cache(
        detector_output_cache_dir, datasets)

    images_to_detect = []
    for img_path in potential_images_to_detect:
        # img_path: <dataset-name>/<img-filename>
        ds, img_file = img_path.split('/', maxsplit=1)
        if img_file not in detection_cache[ds]:
            images_to_detect.append(img_path)

    return images_to_detect, detection_cache, detection_categories


def split_images_list_by_dataset(images_to_detect: Iterable[str]
                                 ) -> dict[str, list[str]]:
    """
    Args:
        images_to_detect: list of str, image paths with the format
            <dataset-name>/<image-filename>

    Returns: dict, maps dataset name to a list of image paths
    """

    images_by_dataset: dict[str, list[str]] = {}
    for img_path in images_to_detect:
        dataset = img_path[:img_path.find('/')]
        if dataset not in images_by_dataset:
            images_by_dataset[dataset] = []
        images_by_dataset[dataset].append(img_path)
    return images_by_dataset


def submit_batch_detection_api(images_to_detect: Iterable[str],
                               task_lists_dir: str,
                               detector_version: str,
                               account: str,
                               container: str,
                               sas_token: str,
                               caller: str,
                               batch_detection_api_url: str,
                               resume_file_path: str
                               ) -> dict[str, list[Task]]:
    """
    Args:
        images_to_detect: list of str, list of str, image paths with the format
            <dataset-name>/<image-filename>
        task_lists_dir: str, path to local directory for saving JSON files
            each containing a list of image URLs corresponding to an API task
        detector_version: str, MegaDetector version string, e.g., '4.1',
            see {batch_detection_api_url}/supported_model_versions
        account: str, Azure Storage account name
        container: str, Azure Blob Storage container name, where the task lists
            will be uploaded
        sas_token: str, SAS token with write permissions for the container
        caller: str, allow-listed caller
        batch_detection_api_url: str, URL to batch detection API
        resume_file_path: str, path to save resume file

    Returns: dict, maps str dataset name to list of Task objects
    """

    filtered_images_to_detect = [
        x for x in images_to_detect if path_utils.is_image_file(x)]
    not_images = set(images_to_detect) - set(filtered_images_to_detect)
    if len(not_images) == 0:
        print('Good! All image files have valid file extensions.')
    else:
        print(f'Skipping {len(not_images)} files with non-image extensions:')
        pprint.pprint(sorted(not_images))
    images_to_detect = filtered_images_to_detect

    datasets_table = megadb_utils.MegadbUtils().get_datasets_table()

    images_by_dataset = split_images_list_by_dataset(images_to_detect)
    tasks_by_dataset = {}
    for dataset, image_paths in images_by_dataset.items():
        # get SAS URL for images container
        images_sas_token = datasets_table[dataset]['container_sas_key']
        if images_sas_token[0] == '?':
            images_sas_token = images_sas_token[1:]
        images_container_url = sas_blob_utils.build_azure_storage_uri(
            account=datasets_table[dataset]['storage_account'],
            container=datasets_table[dataset]['container'],
            sas_token=images_sas_token)

        # strip image paths of dataset name
        image_blob_names = [path[path.find('/') + 1:] for path in image_paths]

        tasks_by_dataset[dataset] = submit_batch_detection_api_by_dataset(
            dataset=dataset,
            image_blob_names=image_blob_names,
            images_container_url=images_container_url,
            task_lists_dir=task_lists_dir,
            detector_version=detector_version,
            account=account, container=container, sas_token=sas_token,
            caller=caller, batch_detection_api_url=batch_detection_api_url)

    # save list of dataset names and task IDs for resuming
    resume_json = [
        {
            'dataset': dataset,
            'task_name': task.name,
            'task_id': task.id,
            'local_images_list_path': task.local_images_list_path
        }
        for dataset in tasks_by_dataset
        for task in tasks_by_dataset[dataset]
    ]
    ct_utils.write_json(resume_file_path, resume_json)
    return tasks_by_dataset


def submit_batch_detection_api_by_dataset(
        dataset: str,
        image_blob_names: Sequence[str],
        images_container_url: str,
        task_lists_dir: str,
        detector_version: str,
        account: str,
        container: str,
        sas_token: str,
        caller: str,
        batch_detection_api_url: str
        ) -> list[Task]:
    """
    Args:
        dataset: str, name of dataset
        image_blob_names: list of str, image blob names from the same dataset
        images_container_url: str, URL to blob storage container where images
            from this dataset are stored, including SAS token with read
            permissions if container is not public
        **see submit_batch_detection_api() for description of other args

    Returns: list of Task objects
    """

    os.makedirs(task_lists_dir, exist_ok=True)

    date = datetime.now().strftime('%Y%m%d_%H%M%S')  # e.g., '20200722_110816'
    task_list_base_filename = f'task_list_{dataset}_{date}.json'

    task_list_paths, _ = divide_list_into_tasks(
        file_list=image_blob_names,
        save_path=os.path.join(task_lists_dir, task_list_base_filename))

    # complete task name: 'detect_for_classifier_caltech_20200722_110816_task01'
    task_name_template = 'detect_for_classifier_{dataset}_{date}_task{n:>02d}'
    tasks: list[Task] = []
    for i, task_list_path in enumerate(task_list_paths):
        task = Task(
            name=task_name_template.format(dataset=dataset, date=date, n=i),
            images_list_path=task_list_path, api_url=batch_detection_api_url)
        task.upload_images_list(
            account=account, container=container, sas_token=sas_token)
        task.generate_api_request(
            caller=caller,
            input_container_url=images_container_url,
            model_version=detector_version)
        print(f'Submitting task for: {task_list_path}')
        task.submit()
        print(f'- task ID: {task.id}')
        tasks.append(task)

        # HACK! Sleep for 10s between task submissions in the hopes that it
        # decreases the chance of backend JSON "database" corruption
        time.sleep(10)
    return tasks


def resume_tasks(resume_file_path: str, batch_detection_api_url: str
                 ) -> dict[str, list[Task]]:
    """
    Args:
        resume_file_path: str, path to resume file with list of info dicts on
            running tasks
        batch_detection_api_url: str, URL to batch detection API

    Returns: dict, maps str dataset name to list of Task objects
    """

    with open(resume_file_path, 'r') as f:
        resume_json = json.load(f)

    tasks_by_dataset: dict[str, list[Task]] = {}
    for info_dict in resume_json:
        dataset = info_dict['dataset']
        if dataset not in tasks_by_dataset:
            tasks_by_dataset[dataset] = []
        task = Task(name=info_dict['task_name'],
                    task_id=info_dict['task_id'],
                    images_list_path=info_dict['local_images_list_path'],
                    validate=False,
                    api_url=batch_detection_api_url)
        tasks_by_dataset[dataset].append(task)
    return tasks_by_dataset


def wait_for_tasks(tasks_by_dataset: Mapping[str, Iterable[Task]],
                   detector_output_cache_dir: str,
                   output_dir: Optional[str] = None,
                   poll_interval: int = 120) -> None:
    """
    Waits for the Batch Detection API tasks to finish running.

    For jobs that finish successfully, merges the output with cached detector
    outputs.

    Args:
        tasks_by_dataset: dict, maps str dataset name to list of Task objects
        detector_output_cache_dir: str, path to local directory where detector
            outputs are cached, 1 JSON file per dataset, directory must
            already exist
        output_dir: optional str, task status responses for completed tasks are
            saved to <output_dir>/batchapi_response/{task_id}.json
        poll_interval: int, # of seconds between pinging the task status API
    """

    remaining_tasks: list[tuple[str, Task]] = [
        (dataset, task) for dataset, tasks in tasks_by_dataset.items()
        for task in tasks]

    progbar = tqdm(total=len(remaining_tasks))
    while True:
        new_remaining_tasks = []
        for dataset, task in remaining_tasks:
            try:
                task.check_status()
            except (BatchAPIResponseError, requests.HTTPError) as e:
                exception_type = type(e).__name__
                tqdm.write(f'Error in checking status of task {task.id}: '
                           f'({exception_type}) {e}')
                tqdm.write(f'Skipping task {task.id}.')
                continue

            # task still running => continue
            if task.status == TaskStatus.RUNNING:
                new_remaining_tasks.append((dataset, task))
                continue

            progbar.update(1)
            tqdm.write(f'Task {task.id} stopped with status {task.status}')

            if task.status in [TaskStatus.PROBLEM, TaskStatus.FAILED]:
                tqdm.write('API response:')
                tqdm.write(str(task.response))
                continue

            # task finished successfully, save response to disk
            assert task.status == TaskStatus.COMPLETED
            if output_dir is not None:
                save_dir = os.path.join(output_dir, 'batchapi_response')
                if not os.path.exists(save_dir):
                    tqdm.write(f'Creating API output dir: {save_dir}')
                    os.makedirs(save_dir)
                ct_utils.write_json(os.path.join(save_dir, f'{task.id}.json'), task.response)
            message = task.response['Status']['message']
            num_failed_shards = message['num_failed_shards']
            if num_failed_shards != 0:
                tqdm.write(f'Task {task.id} completed with {num_failed_shards} '
                           'failed shards.')

            detections_url = message['output_file_urls']['detections']
            if task.id not in detections_url:
                tqdm.write('Invalid detections URL in response. Skipping task.')
                continue

            detections = requests.get(detections_url).json()
            msg = cache_detections(
                detections=detections, dataset=dataset,
                detector_output_cache_dir=detector_output_cache_dir)
            tqdm.write(msg)

        remaining_tasks = new_remaining_tasks
        if len(remaining_tasks) == 0:
            break
        tqdm.write(f'Sleeping for {poll_interval} seconds...')
        time.sleep(poll_interval)

    progbar.close()


def download_and_crop(
        queried_images_json: Mapping[str, Mapping[str, Any]],
        detection_cache: Mapping[str, Mapping[str, Mapping[str, Any]]],
        detection_categories: Mapping[str, str],
        detector_version: str,
        cropped_images_dir: str,
        confidence_threshold: float,
        save_full_images: bool,
        square_crops: bool,
        check_crops_valid: bool,
        images_dir: Optional[str] = None,
        threads: int = 1,
        images_missing_detections: Optional[Iterable[str]] = None
        ) -> tuple[list[str], int, int]:
    """
    Saves crops to a file with the same name as the original image with an
    additional suffix appended, starting with 3 underscores:
    - if image has ground truth bboxes: "___cropXX.jpg", where "XX" indicates
        the bounding box index
    - if image has bboxes from MegaDetector: "___cropXX_mdvY.Y.jpg", where
        "Y.Y" indicates the MegaDetector version
    See module docstring for more info and examples.

    Note: this function is very similar to the "download_and_crop()" function in
        crop_detections.py. The main difference is that this function uses
        MegaDB to look up Azure Storage container information for images based
        on the dataset, whereas the crop_detections.py version has no concept
        of a "dataset" and "ground-truth" bounding boxes from MegaDB.

    Args:
        queried_images_json: dict, represents JSON output of json_validator.py,
            all images in queried_images_json are assumed to have either ground
            truth or cached detected bounding boxes unless
            images_missing_detections is given
        detection_cache: dict, dataset_name => {img_path => detection_dict}
        detector_version: str, detector version string, e.g., '4.1'
        cropped_images_dir: str, path to folder where cropped images are saved
        confidence_threshold: float, only crop bounding boxes above this value
        save_full_images: bool, whether to save downloaded images to images_dir,
            images_dir must be given and must exist if save_full_images=True
        square_crops: bool, whether to crop bounding boxes as squares
        check_crops_valid: bool, whether to load each crop to ensure the file is
            valid (i.e., not truncated)
        images_dir: optional str, path to folder where full images are saved
        threads: int, number of threads to use for downloading images
        images_missing_detections: optional list of str, image files to skip
            because they have no ground truth or cached detected bounding boxes

    Returns: list of str, images with bounding boxes that failed to download or
        crop properly
    """

    # error checking before we download and crop any images
    valid_img_paths = set(queried_images_json.keys())
    if images_missing_detections is not None:
        valid_img_paths -= set(images_missing_detections)
    for img_path in valid_img_paths:
        info_dict = queried_images_json[img_path]
        ds, img_file = img_path.split('/', maxsplit=1)
        assert ds == info_dict['dataset']

        if 'bbox' in info_dict:  # ground-truth bounding boxes
            pass
        elif img_file in detection_cache[ds]:  # detected bounding boxes
            bbox_dicts = detection_cache[ds][img_file]['detections']
            assert all('conf' in bbox_dict for bbox_dict in bbox_dicts)
            # convert from category ID to category name
            for d in bbox_dicts:
                d['category'] = detection_categories[d['category']]
        else:
            raise ValueError(f'{img_path} has no ground truth bounding boxes '
                             'and was not found in the detection cache. Please '
                             'include it in images_missing_detections.')

    # we need the datasets table for getting SAS keys
    datasets_table = megadb_utils.MegadbUtils().get_datasets_table()
    container_clients = {}  # dataset name => ContainerClient

    pool = futures.ThreadPoolExecutor(max_workers=threads)
    future_to_img_path = {}
    images_failed_download = []

    print(f'Getting bbox info for {len(valid_img_paths)} images...')
    for img_path in tqdm(sorted(valid_img_paths)):
        # we already did all error checking above, so we don't do any here
        info_dict = queried_images_json[img_path]
        ds, img_file = img_path.split('/', maxsplit=1)

        # get ContainerClient
        if ds not in container_clients:
            sas_token = datasets_table[ds]['container_sas_key']
            if sas_token[0] == '?':
                sas_token = sas_token[1:]
            url = sas_blob_utils.build_azure_storage_uri(
                account=datasets_table[ds]['storage_account'],
                container=datasets_table[ds]['container'],
                sas_token=sas_token)
            container_clients[ds] = ContainerClient.from_container_url(url)
        container_client = container_clients[ds]

        # get bounding boxes
        # we must include the dataset <ds> in <crop_path_template> because
        #    '{img_path}' actually gets populated with <img_file> in
        #    load_and_crop()
        is_ground_truth = ('bbox' in info_dict)
        if is_ground_truth:  # ground-truth bounding boxes
            bbox_dicts = info_dict['bbox']
            crop_path_template = os.path.join(
                cropped_images_dir, ds, '{img_path}___crop{n:>02d}.jpg')
        else:  # detected bounding boxes
            bbox_dicts = detection_cache[ds][img_file]['detections']
            crop_path_template = os.path.join(
                cropped_images_dir, ds,
                '{img_path}___crop{n:>02d}_' + f'mdv{detector_version}.jpg')

        ds_dir = None if images_dir is None else os.path.join(images_dir, ds)

        # get the image, either from disk or from Blob Storage
        future = pool.submit(
            load_and_crop, img_file, ds_dir, container_client, bbox_dicts,
            confidence_threshold, crop_path_template, save_full_images,
            square_crops, check_crops_valid)
        future_to_img_path[future] = img_path

    total = len(future_to_img_path)
    total_downloads = 0
    total_new_crops = 0
    print(f'Reading/downloading {total} images and cropping...')
    for future in tqdm(futures.as_completed(future_to_img_path), total=total):
        img_path = future_to_img_path[future]
        try:
            did_download, num_new_crops = future.result()
            total_downloads += did_download
            total_new_crops += num_new_crops
        except Exception as e:  # pylint: disable=broad-except
            exception_type = type(e).__name__
            tqdm.write(f'{img_path} - generated {exception_type}: {e}')
            images_failed_download.append(img_path)

    pool.shutdown()
    for container_client in container_clients.values():
        # inelegant way to close the container_clients
        with container_client:
            pass

    print(f'Downloaded {total_downloads} images.')
    print(f'Made {total_new_crops} new crops.')
    return images_failed_download, total_downloads, total_new_crops


#%% Command-line driver

def _parse_args() -> argparse.Namespace:

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Detects and crops images.')
    parser.add_argument(
        'queried_images_json',
        help='path to JSON file mapping image paths and classification info')
    parser.add_argument(
        'output_dir',
        help='path to directory to save log file. If --run-detector, then '
             'task lists and status responses are also saved here.')
    parser.add_argument(
        '-c', '--detector-output-cache-dir', required=True,
        help='(required) path to directory where detector outputs are cached')
    parser.add_argument(
        '-v', '--detector-version', required=True,
        help='(required) detector version string, e.g., "4.1"')
    parser.add_argument(
        '-d', '--run-detector', action='store_true',
        help='Run the Batch Detection API. If not given, skips running the '
             'detector (and only use ground truth and cached bounding boxes).')
    parser.add_argument(
        '-r', '--resume-file',
        help='path to save JSON file with list of info dicts on running tasks, '
             'or to resume from running tasks. Only used if --run-detector is '
             'set. Each dict has keys '
             '["dataset", "task_id", "task_name", "local_images_list_path", '
             '"remote_images_list_url"]')
    parser.add_argument(
        '-p', '--cropped-images-dir',
        help='path to local directory for saving crops of bounding boxes. No '
             'images are downloaded or cropped if this argument is not given.')
    parser.add_argument(
        '--save-full-images', action='store_true',
        help='if downloading an image, save the full image to --images-dir, '
             'only used if <cropped_images_dir> is not None')
    parser.add_argument(
        '--square-crops', action='store_true',
        help='crop bounding boxes as squares, '
             'only used if <cropped_images_dir> is not None')
    parser.add_argument(
        '--check-crops-valid', action='store_true',
        help='load each crop to ensure file is valid (i.e., not truncated), '
             'only used if <cropped_images_dir> is not None')
    parser.add_argument(
        '-t', '--threshold', type=float, default=0.0,
        help='confidence threshold above which to crop bounding boxes, '
             'only used if <cropped_images_dir> is not None')
    parser.add_argument(
        '-i', '--images-dir',
        help='path to local directory where images are saved, '
             'only used if <cropped_images_dir> is not None')
    parser.add_argument(
        '-n', '--threads', type=int, default=1,
        help='number of threads to use for downloading images, '
             'only used if <cropped_images_dir> is not None')
    return parser.parse_args()


if __name__ == '__main__':

    args = _parse_args()
    main(queried_images_json_path=args.queried_images_json,
         output_dir=args.output_dir,
         detector_version=args.detector_version,
         detector_output_cache_base_dir=args.detector_output_cache_dir,
         run_detector=args.run_detector,
         resume_file_path=args.resume_file,
         cropped_images_dir=args.cropped_images_dir,
         save_full_images=args.save_full_images,
         square_crops=args.square_crops,
         check_crops_valid=args.check_crops_valid,
         confidence_threshold=args.threshold,
         images_dir=args.images_dir,
         threads=args.threads)



"""

pytorch_detector.py

Module to run YOLO-based MegaDetector models.

"""

#%% Imports and constants

import os
import sys
import math
import zipfile
import tempfile
import shutil
import uuid
import json
import inspect

import cv2
import torch
import numpy as np

from megadetector.detection.run_detector import \
    CONF_DIGITS, COORD_DIGITS, FAILURE_INFER, FAILURE_IMAGE_OPEN, \
    get_detector_version_from_model_file, \
    known_models
from megadetector.utils.ct_utils import parse_bool_string
from megadetector.utils.ct_utils import is_running_in_gha
from megadetector.utils import ct_utils
import torchvision

# We support a few ways of accessing the YOLOv5 dependencies:
#
# * The standard configuration as of 2023.09 expects that the YOLOv5 repo is checked
#   out and on the PYTHONPATH (import utils)
#
# * Supported but non-default (used for PyPI packaging):
#
#   pip install ultralytics-yolov5
#
# * Works, but not supported:
#
#   pip install yolov5
#
# * Unfinished:
#
#   pip install ultralytics

yolo_model_type_imported = None

def _get_model_type_for_model(model_file,
                              prefer_model_type_source='table',
                              default_model_type='yolov5',
                              verbose=False):
    """
    Determine the model type (i.e., the inference library we need to use) for a .pt file.

    Args:
        model_file (str): the model file to read
        prefer_model_type_source (str, optional): how should we handle the (very unlikely)
            case where the metadata in the file indicates one model type, but the global model
            type table says something else.  Should be "table" (trust the table) or "file"
            (trust the file).
        default_model_type (str, optional): return value for the case where we can't find
            appropriate metadata in the file or in the global table.
        verbose (bool, optional): enable additional debug output

    Returns:
        str: the model type indicated for this model
    """

    model_info = read_metadata_from_megadetector_model_file(model_file)

    # Check whether the model file itself specified a model type
    model_type_from_model_file_metadata = None

    if model_info is not None and 'model_type' in model_info:
        model_type_from_model_file_metadata = model_info['model_type']
        if verbose:
            print('Parsed model type {} from model {}'.format(
                model_type_from_model_file_metadata,
                model_file))

    model_type_from_model_version = None

    # Check whether this is a known model version with a specific model type
    model_version_from_file = get_detector_version_from_model_file(model_file)

    if model_version_from_file is not None and model_version_from_file in known_models:
        model_info = known_models[model_version_from_file]
        if 'model_type' in model_info:
            model_type_from_model_version = model_info['model_type']
            if verbose:
                print('Parsed model type {} from global metadata'.format(model_type_from_model_version))
        else:
            model_type_from_model_version = None

    if model_type_from_model_file_metadata is None and \
        model_type_from_model_version is None:
        if verbose:
            print('Could not determine model type for {}, assuming {}'.format(
                model_file,default_model_type))
        model_type = default_model_type

    elif model_type_from_model_file_metadata is not None and \
         model_type_from_model_version is not None:
        if model_type_from_model_version == model_type_from_model_file_metadata:
            model_type = model_type_from_model_file_metadata
        else:
            print('Warning: model type from model version is {}, from file metadata is {}'.format(
                model_type_from_model_version,model_type_from_model_file_metadata))
            if prefer_model_type_source == 'table':
                model_type = model_type_from_model_file_metadata
            else:
                model_type = model_type_from_model_version

    elif model_type_from_model_file_metadata is not None:

        model_type = model_type_from_model_file_metadata

    elif model_type_from_model_version is not None:

        model_type = model_type_from_model_version

    return model_type

# ...def _get_model_type_for_model(...)


def _initialize_yolo_imports_for_model(model_file,
                                       prefer_model_type_source='table',
                                       default_model_type='yolov5',
                                       detector_options=None,
                                       verbose=False):
    """
    Initialize the appropriate YOLO imports for a model file.

    Args:
        model_file (str): The model file for which we're loading support
        prefer_model_type_source (str, optional): how should we handle the (very unlikely)
            case where the metadata in the file indicates one model type, but the global model
            type table says something else.  Should be "table" (trust the table) or "file"
            (trust the file).
        default_model_type (str, optional): return value for the case where we can't find
            appropriate metadata in the file or in the global table.
        detector_options (dict, optional): dictionary of detector options that mean
            different things to different models
        verbose (bool, optional): enable additional debug output

    Returns:
        str: the model type for which we initialized support
    """

    global yolo_model_type_imported

    if detector_options is not None and 'model_type' in detector_options:
        model_type = detector_options['model_type']
        print('Model type {} provided in detector options'.format(model_type))
    else:
        model_type = _get_model_type_for_model(model_file,
                                               prefer_model_type_source=prefer_model_type_source,
                                               default_model_type=default_model_type)

    if yolo_model_type_imported is not None:
        if model_type == yolo_model_type_imported:
            print('Bypassing imports for model type {}'.format(model_type))
            return
        else:
            print('Previously set up imports for model type {}, re-importing as {}'.format(
                yolo_model_type_imported,model_type))

    _initialize_yolo_imports(model_type,verbose=verbose)

    return model_type


def _clean_yolo_imports(verbose=False, aggressive_cleanup=False):
    """
    Remove all YOLO-related imports from sys.modules and sys.path, to allow a clean re-import
    of another YOLO library version.  The reason we jump through all these hoops, rather than
    just, e.g., handling different libraries in different modules, is that we need to make sure
    *pickle* sees the right version of modules during module loading, including modules we don't
    load directly (i.e., every module loaded within a YOLO library), and the only way I know to
    do that is to remove all the "wrong" versions from sys.modules and sys.path.

    Args:
        verbose (bool, optional): enable additional debug output
        aggressive_cleanup (bool, optional): err on the side of removing modules,
            at least by ignoring whether they are/aren't in a site-packages folder.
            By default, only modules in a folder that includes "site-packages" will
            be considered for unloading.
    """

    modules_to_delete = []

    for module_name in sys.modules.keys():

        module = sys.modules[module_name]
        if not hasattr(module,'__file__') or (module.__file__ is None):
            continue
        try:
            module_file = module.__file__.replace('\\','/')
            if not aggressive_cleanup:
                if 'site-packages' not in module_file:
                    continue
            tokens = module_file.split('/')

            # For local path imports, a module filename that should be unloaded might
            # look like:
            #
            # c:/git/yolov9/models/common.py
            #
            # For pip imports, a module filename that should be unloaded might look like:
            #
            # c:/users/user/miniforge3/envs/megadetector/lib/site-packages/yolov9/utils/__init__.py
            first_token_to_check = len(tokens) - 4
            for i_token,token in enumerate(tokens):
                if i_token < first_token_to_check:
                    continue
                # Don't remove anything based on the environment name, which
                # always follows "envs" in the path
                if (i_token > 1) and (tokens[i_token-1] == 'envs'):
                    continue
                if ('yolov5' in token) or ('yolov9' in token) or ('ultralytics' in token):
                    if verbose:
                        print('Module {} ({}) looks deletable'.format(module_name,module_file))
                    modules_to_delete.append(module_name)
                    break
        except Exception as e:
            if verbose:
                print('Exception during module review: {}'.format(str(e)))
            pass

    # ...for each module in the global namespace

    for module_name in modules_to_delete:

        if module_name in sys.modules.keys():
            if verbose:
                try:
                    module = sys.modules[module_name]
                    module_file = module.__file__.replace('\\','/')
                    print('clean_yolo_imports: deleting module {}: {}'.format(module_name,module_file))
                except Exception:
                    pass
            del sys.modules[module_name]

    # ...for each module we want to remove from the global namespace

    paths_to_delete = []

    for p in sys.path:
        if p.endswith('yolov5') or p.endswith('yolov9') or p.endswith('ultralytics'):
            print('clean_yolo_imports: removing {} from path'.format(p))
            paths_to_delete.append(p)

    for p in paths_to_delete:
        sys.path.remove(p)

# ...def _clean_yolo_imports(...)


def _initialize_yolo_imports(model_type='yolov5',
                             allow_fallback_import=True,
                             force_reimport=False,
                             verbose=False):
    """
    Imports required functions from one or more yolo libraries (yolov5, yolov9,
    ultralytics, targeting support for [model_type]).

    Args:
        model_type (str): The model type for which we're loading support
        allow_fallback_import (bool, optional): If we can't import from the package for
            which we're trying to load support, fall back to "import utils".  This is
            typically used when the right support library is on the current PYTHONPATH.
        force_reimport (bool, optional): import the appropriate libraries even if the
            requested model type matches the current initialization state
        verbose (bool, optional): include additional debug output

    Returns:
        str: the model type for which we initialized support
    """

    # When running in pytest, the megadetector 'utils' module is put in the global
    # namespace, which creates conflicts with yolov5; remove it from the global
    # namespsace.
    if ('PYTEST_CURRENT_TEST' in os.environ):
        print('*** pytest detected ***')
        if ('utils' in sys.modules):
            utils_module = sys.modules['utils']
            if hasattr(utils_module, '__file__') and 'megadetector' in str(utils_module.__file__):
                print(f"Removing conflicting utils module: {utils_module.__file__}")
                sys.modules.pop('utils', None)
                # Also remove any submodules
                to_remove = [name for name in sys.modules if name.startswith('utils.')]
                for name in to_remove:
                    sys.modules.pop(name, None)

    global yolo_model_type_imported

    if model_type is None:
        model_type = 'yolov5'

    # The point of this function is to make the appropriate version
    # of the following functions available at module scope
    global non_max_suppression
    global xyxy2xywh
    global letterbox
    global scale_coords

    if yolo_model_type_imported is not None:
        if (yolo_model_type_imported == model_type) and (not force_reimport):
            print('Bypassing imports for YOLO model type {}'.format(model_type))
            return
        else:
            _clean_yolo_imports()

    try_yolov5_import = (model_type == 'yolov5')
    try_yolov9_import = (model_type == 'yolov9')
    try_ultralytics_import = (model_type == 'ultralytics')

    utils_imported = False

    # First try importing from the yolov5 package; this is how the pip
    # package finds YOLOv5 utilities.
    if try_yolov5_import and not utils_imported:

        try:
            from yolov5.utils.general import non_max_suppression, xyxy2xywh # noqa
            from yolov5.utils.augmentations import letterbox # noqa
            try:
                from yolov5.utils.general import scale_boxes as scale_coords
            except Exception:
                from yolov5.utils.general import scale_coords
            utils_imported = True
            if verbose:
                print('Imported utils from YOLOv5 package')

        except Exception as e: # noqa
            # print('yolov5 module import failed: {}'.format(e))
            # print(traceback.format_exc())
            pass

    # Next try importing from the yolov9 package
    if try_yolov9_import and not utils_imported:

        try:

            from yolov9.utils.general import non_max_suppression, xyxy2xywh # noqa
            from yolov9.utils.augmentations import letterbox # noqa
            from yolov9.utils.general import scale_boxes as scale_coords # noqa
            utils_imported = True
            if verbose:
                print('Imported utils from YOLOv9 package')

        except Exception as e: # noqa

            # print('yolov9 module import failed: {}'.format(e))
            # print(traceback.format_exc())
            pass

    # If we haven't succeeded yet, import from the ultralytics package
    if try_ultralytics_import and not utils_imported:

        try:

            import ultralytics # type: ignore # noqa

        except Exception:

            print('It looks like you are trying to run a model that requires the ultralytics package, '
                  'but the ultralytics package is not installed, but .  For licensing reasons, this '
                  'is not installed by default with the MegaDetector Python package.  Run '
                  '"pip install ultralytics" to install it, and try again.')
            raise

        try:

            from ultralytics.utils.ops import non_max_suppression # type: ignore # noqa
            from ultralytics.utils.ops import xyxy2xywh # type: ignore # noqa

            # In the ultralytics package, scale_boxes and scale_coords both exist;
            # we want scale_boxes.
            #
            # from ultralytics.utils.ops import scale_coords # noqa
            from ultralytics.utils.ops import scale_boxes as scale_coords # type: ignore # noqa
            from ultralytics.data.augment import LetterBox # type: ignore # noqa

            # letterbox() became a LetterBox class in the ultralytics package.  Create a
            # backwards-compatible letterbox function wrapper that wraps the class up.
            def letterbox(img,new_shape,auto=False,scaleFill=False, #noqa
                          scaleup=True,center=True,stride=32):

                # Ultralytics changed the "scaleFill" parameter to "scale_fill", we want to support
                # both conventions.
                use_old_scalefill_arg = False
                try:
                    sig = inspect.signature(LetterBox.__init__)
                    if 'scaleFill' in sig.parameters:
                        use_old_scalefill_arg = True
                except Exception:
                    pass

                if use_old_scalefill_arg:
                    if verbose:
                        print('Using old scaleFill calling convention')
                    letterbox_transformer = LetterBox(new_shape,auto=auto,scaleFill=scaleFill,
                                                    scaleup=scaleup,center=center,stride=stride)
                else:
                    letterbox_transformer = LetterBox(new_shape,auto=auto,scale_fill=scaleFill,
                                                  scaleup=scaleup,center=center,stride=stride)

                letterbox_result = letterbox_transformer(image=img)

                if isinstance(new_shape,int):
                    new_shape = [new_shape,new_shape]

                # The letterboxing is done, we just need to reverse-engineer what it did
                shape = img.shape[:2]

                r = min(new_shape[0] / shape[0], new_shape[1] / shape[1])
                if not scaleup:
                    r = min(r, 1.0)
                ratio = r, r

                new_unpad = int(round(shape[1] * r)), int(round(shape[0] * r))
                dw, dh = new_shape[1] - new_unpad[0], new_shape[0] - new_unpad[1]
                if auto:
                    dw, dh = np.mod(dw, stride), np.mod(dh, stride)
                elif scaleFill:
                    dw, dh = 0.0, 0.0
                    new_unpad = (new_shape[1], new_shape[0])
                    ratio = (new_shape[1] / shape[1], new_shape[0] / shape[0])

                dw /= 2
                dh /= 2
                pad = (dw,dh)

                return [letterbox_result,ratio,pad]

            utils_imported = True
            if verbose:
                print('Imported utils from ultralytics package')

        except Exception:

            # print('Ultralytics module import failed')
            pass

    # If we haven't succeeded yet, assume the YOLOv5 repo is on our PYTHONPATH.
    if (not utils_imported) and allow_fallback_import:

        try:

            # import pre- and post-processing functions from the YOLOv5 repo
            from utils.general import non_max_suppression, xyxy2xywh # type: ignore
            from utils.augmentations import letterbox # type: ignore

            # scale_coords() is scale_boxes() in some YOLOv5 versions
            try:
                from utils.general import scale_coords # type: ignore
            except ImportError:
                from utils.general import scale_boxes as scale_coords # type: ignore
            utils_imported = True
            imported_file = sys.modules[scale_coords.__module__].__file__
            if verbose:
                print('Imported utils from {}'.format(imported_file))

        except ModuleNotFoundError as e:

            raise ModuleNotFoundError('Could not import YOLOv5 functions:\n{}'.format(str(e)))

    assert utils_imported, 'YOLO utils import error'

    yolo_model_type_imported = model_type
    if verbose:
        print('Prepared YOLO imports for model type {}'.format(model_type))

    return model_type

# ...def _initialize_yolo_imports(...)


#%% NMS

def nms(prediction, conf_thres=0.25, iou_thres=0.45, max_det=300):
    """
    Non-maximum suppression (a wrapper around torchvision.ops.nms())

    Args:
        prediction (torch.Tensor): Model predictions with shape [batch_size, num_anchors, num_classes + 5]
            Format: [x_center, y_center, width, height, objectness, class1_conf, class2_conf, ...]
            Coordinates are normalized to input image size.
        conf_thres (float): Confidence threshold for filtering detections
        iou_thres (float): IoU threshold for NMS
        max_det (int): Maximum number of detections per image

    Returns:
        list: List of tensors, one per image in batch. Each tensor has shape [N, 6] where:
                - N is the number of detections for that image
                - Columns are [x1, y1, x2, y2, confidence, class_id]
                - Coordinates are in absolute pixels relative to input image size
                - class_id is the integer class index (0-based)
    """

    batch_size = prediction.shape[0]
    num_classes = prediction.shape[2] - 5 # noqa
    output = []

    # Process each image in the batch
    for img_idx in range(batch_size):

        x = prediction[img_idx]  # Shape: [num_anchors, num_classes + 5]

        # Filter by objectness confidence
        obj_conf = x[:, 4]
        valid_detections = obj_conf > conf_thres
        x = x[valid_detections]

        if x.shape[0] == 0:
            # No detections for this image
            output.append(torch.zeros((0, 6), device=prediction.device))
            continue

        # Convert box coordinates from [x_center, y_center, w, h] to [x1, y1, x2, y2]
        box = x[:, :4].clone()
        box[:, 0] = x[:, 0] - x[:, 2] / 2.0  # x1 = center_x - width/2
        box[:, 1] = x[:, 1] - x[:, 3] / 2.0  # y1 = center_y - height/2
        box[:, 2] = x[:, 0] + x[:, 2] / 2.0  # x2 = center_x + width/2
        box[:, 3] = x[:, 1] + x[:, 3] / 2.0  # y2 = center_y + height/2

        # Get class predictions: multiply objectness by class probabilities
        class_conf = x[:, 5:] * x[:, 4:5]  # shape: [N, num_classes]

        # For each detection, take the class with highest confidence (single-label)
        best_class_conf, best_class_idx = class_conf.max(1, keepdim=True)

        # Filter by class confidence threshold
        conf_mask = best_class_conf.view(-1) > conf_thres
        if conf_mask.sum() == 0:
            # No detections pass confidence threshold
            output.append(torch.zeros((0, 6), device=prediction.device))
            continue

        box = box[conf_mask]
        best_class_conf = best_class_conf[conf_mask]
        best_class_idx = best_class_idx[conf_mask]

        # Prepare for NMS: group detections by class
        unique_classes = best_class_idx.unique()
        final_detections = []

        for class_id in unique_classes:

            class_mask = (best_class_idx == class_id).view(-1)
            class_boxes = box[class_mask]
            class_scores = best_class_conf[class_mask].view(-1)

            if class_boxes.shape[0] == 0:
                continue

            # Apply NMS for this class
            keep_indices = torchvision.ops.nms(class_boxes, class_scores, iou_thres)

            if len(keep_indices) > 0:
                kept_boxes = class_boxes[keep_indices]
                kept_scores = class_scores[keep_indices]
                kept_classes = torch.full((len(keep_indices), 1), class_id.item(),
                                        device=prediction.device, dtype=torch.float)

                # Combine: [x1, y1, x2, y2, conf, class]
                class_detections = torch.cat([kept_boxes, kept_scores.unsqueeze(1), kept_classes], 1)
                final_detections.append(class_detections)

        # ...for each category

        if final_detections:

            # Combine all classes and sort by confidence
            all_detections = torch.cat(final_detections, 0)
            conf_sort_indices = all_detections[:, 4].argsort(descending=True)
            all_detections = all_detections[conf_sort_indices]

            # Limit to max_det
            if all_detections.shape[0] > max_det:
                all_detections = all_detections[:max_det]

            output.append(all_detections)
        else:
            output.append(torch.zeros((0, 6), device=prediction.device))

    # ...for each image in the batch

    return output

# ...def nms(...)


#%% Model metadata functions

def add_metadata_to_megadetector_model_file(model_file_in,
                                            model_file_out,
                                            metadata,
                                            destination_path='megadetector_info.json'):
    """
    Adds a .json file to the specified MegaDetector model file containing metadata used
    by this module.  Always over-writes the output file.

    Args:
        model_file_in (str): The input model filename, typically .pt (.zip is also sensible)
        model_file_out (str): The output model filename, typically .pt (.zip is also sensible).
            May be the same as model_file_in.
        metadata (dict): The metadata dict to add to the output model file
        destination_path (str, optional): The relative path within the main folder of the
            model archive where we should write the metadata.  This is not relative to the root
            of the archive, it's relative to the one and only folder at the root of the archive
            (this is a PyTorch convention).
    """

    tmp_base = os.path.join(tempfile.gettempdir(),'md_metadata')
    os.makedirs(tmp_base,exist_ok=True)
    metadata_tmp_file_relative = 'megadetector_info_' + str(uuid.uuid1()) + '.json'
    metadata_tmp_file_abs = os.path.join(tmp_base,metadata_tmp_file_relative)

    with open(metadata_tmp_file_abs,'w') as f:
        json.dump(metadata,f,indent=1)

    # Copy the input file to the output file
    shutil.copyfile(model_file_in,model_file_out)

    # Write metadata to the output file
    with zipfile.ZipFile(model_file_out, 'a', compression=zipfile.ZIP_DEFLATED) as zipf:

        # Torch doesn't like anything in the root folder of the zipfile, so we put
        # it in the one and only folder.
        names = zipf.namelist()
        root_folders = set()
        for name in names:
            root_folder = name.split('/')[0]
            root_folders.add(root_folder)
        assert len(root_folders) == 1,\
            'This archive does not have exactly one folder at the top level; are you sure it\'s a Torch model file?'
        root_folder = next(iter(root_folders))

        zipf.write(metadata_tmp_file_abs,
                   root_folder + '/' + destination_path,
                   compresslevel=9,
                   compress_type=zipfile.ZIP_DEFLATED)

    try:
        os.remove(metadata_tmp_file_abs)
    except Exception as e:
        print('Warning: error deleting file {}: {}'.format(metadata_tmp_file_abs,str(e)))

# ...def add_metadata_to_megadetector_model_file(...)


def read_metadata_from_megadetector_model_file(model_file,
                                               relative_path='megadetector_info.json',
                                               verbose=False):
    """
    Reads custom MegaDetector metadata from a modified MegaDetector model file.

    Args:
        model_file (str): The model filename to read, typically .pt (.zip is also sensible)
        relative_path (str, optional): The relative path within the main folder of the model
            archive from which we should read the metadata.  This is not relative to the root
            of the archive, it's relative to the one and only folder at the root of the archive
            (this is a PyTorch convention).
        verbose (str, optional): enable additional debug output

    Returns:
        object: whatever we read from the metadata file, always a dict in practice.  Returns
        None if we failed to read the specified metadata file.
    """

    with zipfile.ZipFile(model_file,'r') as zipf:

        # Torch doesn't like anything in the root folder of the zipfile, so we put
        # it in the one and only folder.
        names = zipf.namelist()
        root_folders = set()
        for name in names:
            root_folder = name.split('/')[0]
            root_folders.add(root_folder)
        if len(root_folders) != 1:
            print('Warning: this archive does not have exactly one folder at the top level; ' + \
                  'are you sure it\'s a Torch model file?')
            return None
        root_folder = next(iter(root_folders))

        metadata_file = root_folder + '/' + relative_path
        if metadata_file not in names:
            # This is the case for MDv5a and MDv5b
            if verbose:
                print('Warning: could not find metadata file {} in zip archive {}'.format(
                    metadata_file,os.path.basename(model_file)))
            return None

        try:
            path = zipfile.Path(zipf,metadata_file)
            contents = path.read_text()
            d = json.loads(contents)
        except Exception as e:
            print('Warning: error reading metadata from path {}: {}'.format(metadata_file,str(e)))
            return None

        return d

    # ...with zipfile.Zipfile(...)

# ...def read_metadata_from_megadetector_model_file(...)


#%% Inference classes

default_compatibility_mode = 'classic'

# This is a useful hack when I want to verify that my test driver (md_tests.py) is
# correctly forcing a specific compatibility mode (I use "classic-test" in that case)
require_non_default_compatibility_mode = False

class PTDetector:
    """
    Class that runs a PyTorch-based MegaDetector model.  Also used as a preprocessor
    for images that will later be run through an instance of PTDetector.
    """

    def __init__(self, model_path, detector_options=None, verbose=False):
        """
        PTDetector constructor.  If detector_options['preprocess_only'] exists and is
        True, this instance is being used as a preprocessor, so we don't load model weights.
        """

        if verbose:
            print('Initializing PTDetector (verbose)')

        # Set up the import environment for this model, unloading previous
        # YOLO library versions if necessary.
        _initialize_yolo_imports_for_model(model_path,
                                           detector_options=detector_options,
                                           verbose=verbose)

        # Parse options specific to this detector family
        force_cpu = False
        use_model_native_classes = False
        compatibility_mode = default_compatibility_mode

        if detector_options is not None:

            if 'force_cpu' in detector_options:
                force_cpu = parse_bool_string(detector_options['force_cpu'])
            if 'use_model_native_classes' in detector_options:
                use_model_native_classes = parse_bool_string(detector_options['use_model_native_classes'])
            if 'compatibility_mode' in detector_options:
                if detector_options['compatibility_mode'] is None:
                    compatibility_mode = default_compatibility_mode
                else:
                    compatibility_mode = detector_options['compatibility_mode']

        # This is a global option used only during testing, to make sure I'm hitting
        # the cases where we are not using "classic" preprocessing.
        if require_non_default_compatibility_mode:

            print('### DEBUG: requiring non-default compatibility mode ###')
            assert compatibility_mode != 'classic'
            assert compatibility_mode != 'default'

        preprocess_only = False
        if (detector_options is not None) and \
           ('preprocess_only' in detector_options) and \
           (detector_options['preprocess_only']):
            preprocess_only = True

        if verbose or (not preprocess_only):
            print('Loading PT detector with compatibility mode {}'.format(compatibility_mode))

        self.model_metadata = read_metadata_from_megadetector_model_file(model_path)

        #: Image size passed to the letterbox() function; 1280 means "1280 on the long side,
        #: preserving aspect ratio".
        if self.model_metadata is not None and 'image_size' in self.model_metadata:
            self.default_image_size = self.model_metadata['image_size']
            print('Loaded image size {} from model metadata'.format(self.default_image_size))
        else:
            # This is not the default for most YOLO models, but most of the time, if someone
            # is loading a model here that does not have metadata, it's MDv5[ab].0.0
            print('No image size available in model metadata, defaulting to 1280')
            self.default_image_size = 1280

        #: Either a string ('cpu','cuda:0') or a torch.device()
        self.device = 'cpu'

        #: Have we already printed a warning about using a non-standard image size?
        #:
        #: :meta private:
        self.printed_image_size_warning = False

        #: If this is False, we assume the underlying model is producing class indices in the
        #: set (0,1,2) (and we assert() on this), and we add 1 to get to the backwards-compatible
        #: MD classes (1,2,3) before generating output.  If this is True, we use whatever
        #: indices the model provides
        self.use_model_native_classes = use_model_native_classes

        #: This allows us to maintain backwards compatibility across a set of changes to the
        #: way this class does inference.  Currently should start with either "default" or
        #: "classic".
        self.compatibility_mode = compatibility_mode

        #: Stride size passed to the YOLO letterbox() function
        self.letterbox_stride = 32

        # This is a convenient heuristic to determine the stride size without actually loading
        # the model: the only models in the YOLO family with a stride size of 64 are the
        # YOLOv5*6 and YOLOv5*6u models, which are 1280px models.
        #
        # See:
        #
        # github.com/ultralytics/ultralytics/issues/21544
        #
        # Note to self, though, if I decide later to require loading the model on preprocessing
        # workers so I can more reliably choose a stride, this is the right way to determine the
        # stride:
        #
        # self.letterbox_stride = int(self.model.stride.max())
        if self.default_image_size == 1280:
            self.letterbox_stride = 64

        print('Using model stride: {}'.format(self.letterbox_stride))

        #: Use half-precision inference... fixed by the model, generally don't mess with this
        self.half_precision = False

        if preprocess_only:
            return

        if not force_cpu:
            if torch.cuda.is_available():
                self.device = torch.device('cuda:0')
            try:
                if torch.backends.mps.is_built and torch.backends.mps.is_available():
                    # MPS inference fails on GitHub runners as of 2025.08.  This is
                    # independent of model size.  So, we disable MPS when running in GHA.
                    if is_running_in_gha():
                        print('GitHub actions detected, bypassing MPS backend')
                    else:
                        print('Using MPS device')
                        self.device = 'mps'
            except AttributeError:
                pass

        try:
            self.model = PTDetector._load_model(model_path,
                                                device=self.device,
                                                compatibility_mode=self.compatibility_mode)

        except Exception as e:
            # In a very esoteric scenario where an old version of YOLOv5 is used to run
            # newer models, we run into an issue because the "Model" class became
            # "DetectionModel".  New YOLOv5 code handles this case by just setting them
            # to be the same, so doing that externally doesn't seem *that* rude.
            if "Can't get attribute 'DetectionModel'" in str(e):
                print('Forward-compatibility issue detected, patching')
                from models import yolo # type: ignore
                yolo.DetectionModel = yolo.Model
                self.model = PTDetector._load_model(model_path,
                                                    device=self.device,
                                                    compatibility_mode=self.compatibility_mode,
                                                    verbose=verbose)
            else:
                raise
        if (self.device != 'cpu'):
            if verbose:
                print('Sending model to GPU')
            self.model.to(self.device)


    @staticmethod
    def _load_model(model_pt_path, device, compatibility_mode='', verbose=False):

        if verbose:
            print(f'Using PyTorch version {torch.__version__}')

        # I get quirky errors when loading YOLOv5 models on MPS hardware using
        # map_location, but this is the recommended method, so I'm using it everywhere
        # other than MPS devices.
        use_map_location = (device != 'mps')

        if use_map_location:
            try:
                checkpoint = torch.load(model_pt_path, map_location=device, weights_only=False)
            # For a transitional period, we want to support torch 1.1x, where the weights_only
            # parameter doesn't exist
            except Exception as e:
                if "'weights_only' is an invalid keyword" in str(e):
                    checkpoint = torch.load(model_pt_path, map_location=device)
                else:
                    raise
        else:
            try:
                checkpoint = torch.load(model_pt_path, weights_only=False)
            # For a transitional period, we want to support torch 1.1x, where the weights_only
            # parameter doesn't exist
            except Exception as e:
                if "'weights_only' is an invalid keyword" in str(e):
                    checkpoint = torch.load(model_pt_path)
                else:
                    raise

        # Compatibility fix that allows us to load older YOLOv5 models with
        # newer versions of YOLOv5/PT
        for m in checkpoint['model'].modules():
            t = type(m)
            if t is torch.nn.Upsample and not hasattr(m, 'recompute_scale_factor'):
                m.recompute_scale_factor = None

        # Calling .to(device) should no longer be necessary now that we're using map_location=device
        # model = checkpoint['model'].float().fuse().eval().to(device)
        model = checkpoint['model'].float().fuse().eval()

        return model

    # ...def _load_model(...)


    def preprocess_image(self,
                         img_original,
                         image_id='unknown',
                         image_size=None,
                         verbose=False):
        """
        Prepare an image for detection, including scaling and letterboxing.

        Args:
            img_original (Image or np.array): the image on which we should run the detector, with
                EXIF rotation already handled
            image_id (str, optional): a path to identify the image; will be in the "file" field
                of the output object
            detection_threshold (float, optional): only detections above this confidence threshold
                will be included in the return value
            image_size (int, optional): image size (long side) to use for inference, or None to
                use the default size specified at the time the model was loaded
            verbose (bool, optional): enable additional debug output

        Returns:
            dict: dict with fields:
                - file (filename)
                - img (the preprocessed np.array)
                - img_original (the input image before preprocessing, as an np.array)
                - img_original_pil (the input image before preprocessing, as a PIL Image)
                - target_shape (the 2D shape to which the image was resized during preprocessing)
                - scaling_shape (the 2D original size, for normalizing coordinates later)
                - letterbox_ratio (letterbox parameter used for normalizing coordinates later)
                - letterbox_pad (letterbox parameter used for normalizing coordinates later)
        """

        # Prepare return dict
        result = {'file': image_id }

        # Store the PIL version of the original image, the caller may want to use
        # it for metadata extraction later.
        img_original_pil = None

        # If we were given a PIL image, rather than a numpy array
        if not isinstance(img_original,np.ndarray):
            img_original_pil = img_original
            img_original = np.asarray(img_original)

        # PIL images are RGB already
        # img_original = img_original[:, :, ::-1]

        # Save the original shape for scaling boxes later
        scaling_shape = img_original.shape

        # If the caller is requesting a specific target size...
        if image_size is not None:

            assert isinstance(image_size,int)

            if not self.printed_image_size_warning:
                print('Using user-supplied image size {}'.format(image_size))
                self.printed_image_size_warning = True

        # Otherwise resize to self.default_image_size
        else:

            image_size = self.default_image_size
            self.printed_image_size_warning = False

        # ...if the caller has specified an image size

        # In "classic mode", we only do the letterboxing resize, we don't do an
        # additional initial resizing operation
        if 'classic' in self.compatibility_mode:

            resize_ratio = 1.0

        # Resize the image so the long side matches the target image size.  This is not
        # letterboxing (i.e., padding) yet, just resizing.
        else:

            use_ceil_for_resize = ('use_ceil_for_resize' in self.compatibility_mode)

            h,w = img_original.shape[:2]
            resize_ratio = image_size / max(h,w)

            # Only resize if we have to
            if resize_ratio != 1:

                # Match what yolov5 does: use linear interpolation for upsizing;
                # area interpolation for downsizing
                if resize_ratio > 1:
                    interpolation_method = cv2.INTER_LINEAR
                else:
                    interpolation_method = cv2.INTER_AREA

                if use_ceil_for_resize:
                    target_w = math.ceil(w * resize_ratio)
                    target_h = math.ceil(h * resize_ratio)
                else:
                    target_w = int(w * resize_ratio)
                    target_h = int(h * resize_ratio)

                img_original = cv2.resize(
                    img_original, (target_w, target_h),
                    interpolation=interpolation_method)

        if 'classic' in self.compatibility_mode:

            letterbox_auto = True
            letterbox_scaleup = True
            target_shape = image_size

        else:

            letterbox_auto = False
            letterbox_scaleup = False

            # The padding to apply as a fraction of the stride size
            pad = 0.5

            # Resize to a multiple of the model stride
            #
            # This is how we would determine the stride if we knew the model had been loaded:
            #
            # model_stride = int(self.model.stride.max())
            #
            # ...but because we do this on preprocessing workers now, we try to avoid loading the model
            # just for preprocessing, and we assume the stride was determined at the time the PTDetector
            # object was created.
            try:
                model_stride = int(self.model.stride.max())
                if model_stride != self.letterbox_stride:
                    print('*** Warning: model stride is {}, stride at construction time was {} ***'.format(
                        model_stride,self.letterbox_stride
                    ))
            except Exception:
                pass

            model_stride = self.letterbox_stride
            max_dimension = max(img_original.shape)
            normalized_shape = [img_original.shape[0] / max_dimension,
                                img_original.shape[1] / max_dimension]
            target_shape = np.ceil(((np.array(normalized_shape) * image_size) / model_stride) + \
                                   pad).astype(int) * model_stride

        # Now we letterbox, which is just padding, since we've already resized
        img,letterbox_ratio,letterbox_pad = letterbox(img_original,
                                                      new_shape=target_shape,
                                                      stride=self.letterbox_stride,
                                                      auto=letterbox_auto,
                                                      scaleFill=False,
                                                      scaleup=letterbox_scaleup)

        result['img_processed'] = img
        result['img_original'] = img_original
        result['img_original_pil'] = img_original_pil
        result['target_shape'] = target_shape
        result['scaling_shape'] = scaling_shape
        result['letterbox_ratio'] = letterbox_ratio
        result['letterbox_pad'] = letterbox_pad
        return result

    # ...def preprocess_image(...)


    def generate_detections_one_batch(self,
                                      img_original,
                                      image_id=None,
                                      detection_threshold=0.00001,
                                      image_size=None,
                                      augment=False,
                                      verbose=False):
        """
        Run a detector on a batch of images.

        Args:
            img_original (list): list of images (Image, np.array, or dict) on which we should run the detector, with
                EXIF rotation already handled, or dicts representing preprocessed images with associated
                letterbox parameters
            image_id (list or None): list of paths to identify the images; will be in the "file" field
                of the output objects. Will be ignored when img_original contains preprocessed dicts.
            detection_threshold (float, optional): only detections above this confidence threshold
                will be included in the return value
            image_size (int, optional): image size (long side) to use for inference, or None to
                use the default size specified at the time the model was loaded
            augment (bool, optional): enable (implementation-specific) image augmentation
            verbose (bool, optional): enable additional debug output

        Returns:
            list: a list of dictionaries, each with the following fields:
                - 'file' (filename, always present)
                - 'max_detection_conf' (removed from MegaDetector output files by default, but generated here)
                - 'detections' (a list of detection objects containing keys 'category', 'conf', and 'bbox')
                - 'failure' (a failure string, or None if everything went fine)
        """

        # Validate inputs
        if not isinstance(img_original, list):
            raise ValueError('img_original must be a list for batch processing')

        if len(img_original) == 0:
            return []

        # Check input consistency
        if isinstance(img_original[0], dict):
            # All items in img_original should be preprocessed dicts
            for i, img in enumerate(img_original):
                if not isinstance(img, dict):
                    raise ValueError(f'Mixed input types in batch: item {i} is not a dict, but item 0 is a dict')
        else:
            # All items in img_original should be PIL/numpy images, and image_id should be a list of strings
            if image_id is None:
                raise ValueError('image_id must be a list when img_original contains PIL/numpy images')
            if not isinstance(image_id, list):
                raise ValueError('image_id must be a list for batch processing')
            if len(image_id) != len(img_original):
                raise ValueError(
                    'Length mismatch: img_original has {} items, image_id has {} items'.format(
                    len(img_original),len(image_id)))
            for i_img, img in enumerate(img_original):
                if isinstance(img, dict):
                    raise ValueError(
                        'Mixed input types in batch: item {} is a dict, but item 0 is not a dict'.format(
                            i_img))

        if detection_threshold is None:
            detection_threshold = 0.0

        batch_size = len(img_original)
        results = [None] * batch_size

        # Preprocess all images, handling failures
        preprocessed_images = []
        preprocessing_failed_indices = set()

        for i_img, img in enumerate(img_original):

            try:
                if isinstance(img, dict):
                    # Already preprocessed
                    image_info = img
                    current_image_id = image_info['file']
                else:
                    # Need to preprocess
                    current_image_id = image_id[i_img]
                    image_info = self.preprocess_image(
                        img_original=img,
                        image_id=current_image_id,
                        image_size=image_size,
                        verbose=verbose)

                preprocessed_images.append((i_img, image_info, current_image_id))

            except Exception as e:
                print('Warning: preprocessing failed for image {}: {}'.format(
                    image_id[i_img] if image_id else f'index_{i_img}', str(e)))

                preprocessing_failed_indices.add(i_img)
                current_image_id = image_id[i_img] if image_id else f'index_{i_img}'
                results[i_img] = {
                    'file': current_image_id,
                    'detections': None,
                    'failure': FAILURE_IMAGE_OPEN
                }

        # ...for each image in this batch

        # Group preprocessed images by actual processed image shape for batching
        shape_groups = {}
        for original_idx, image_info, current_image_id in preprocessed_images:
            # Use the actual processed image shape for grouping, not target_shape
            actual_shape = tuple(image_info['img_processed'].shape)
            if actual_shape not in shape_groups:
                shape_groups[actual_shape] = []
            shape_groups[actual_shape].append((original_idx, image_info, current_image_id))

        # Process each shape group as a batch
        for target_shape, group_items in shape_groups.items():

            try:
                self._process_batch_group(group_items, results, detection_threshold, augment, verbose)
            except Exception as e:
                # If inference fails for the entire batch, mark all images in this batch as failed
                print('Warning: batch inference failed for shape {}: {}'.format(target_shape, str(e)))

                for original_idx, image_info, current_image_id in group_items:
                    results[original_idx] = {
                        'file': current_image_id,
                        'detections': None,
                        'failure': FAILURE_INFER
                    }

        # ...for each shape group
        return results

    # ...def generate_detections_one_batch(...)


    def _process_batch_group(self, group_items, results, detection_threshold, augment, verbose):
        """
        Process a group of images with the same target shape as a single batch.

        Args:
            group_items (list): List of (original_idx, image_info, current_image_id) tuples
            results (list): Results list to populate (modified in place)
            detection_threshold (float): Detection confidence threshold
            augment (bool): Enable augmentation
            verbose (bool): Enable verbose output

        Returns:
            list of dict: list of dictionaries the same length as group_items, with fields 'file',
            'detections', 'max_detection_conf'.
        """

        if len(group_items) == 0:
            return

        # Extract batch data
        batch_images = []
        batch_metadata = []

        # For each image in this batch...
        for original_idx, image_info, current_image_id in group_items:

            img = image_info['img_processed']

            # Convert HWC to CHW and prepare tensor
            img_tensor = img.transpose((2, 0, 1))
            img_tensor = np.ascontiguousarray(img_tensor)
            img_tensor = torch.from_numpy(img_tensor)
            batch_images.append(img_tensor)

            metadata = {
                'original_idx': original_idx,
                'current_image_id': current_image_id,
                'scaling_shape': image_info['scaling_shape'],
                'letterbox_pad': image_info['letterbox_pad'],
                'img_original': image_info['img_original']
            }
            batch_metadata.append(metadata)

        # ...for each image in this batch

        # Stack images into a batch tensor
        batch_tensor = torch.stack(batch_images)

        batch_tensor = batch_tensor.float()
        batch_tensor /= 255.0

        batch_tensor = batch_tensor.to(self.device)
        if self.half_precision:
            batch_tensor = batch_tensor.half()

        # Run the model on the batch
        pred = self.model(batch_tensor, augment=augment)[0]

        # Configure NMS parameters
        if 'classic' in self.compatibility_mode:
            nms_iou_thres = 0.45
        else:
            nms_iou_thres = 0.6

        pred = nms(prediction=pred,
                   conf_thres=detection_threshold,
                   iou_thres=nms_iou_thres)

        # For posterity, the ultralytics implementation
        if False:
            pred = non_max_suppression(prediction=pred,
                                       conf_thres=detection_threshold,
                                       iou_thres=nms_iou_thres,
                                       agnostic=False,
                                       multi_label=False)

        assert isinstance(pred, list)
        assert len(pred) == len(batch_metadata), \
            print('Mismatch between prediction length {} and batch size {}'.format(
                   len(pred),len(batch_metadata)))

        # Process each image's detections
        for i_image, det in enumerate(pred):

            metadata = batch_metadata[i_image]
            original_idx = metadata['original_idx']
            current_image_id = metadata['current_image_id']
            scaling_shape = metadata['scaling_shape']
            letterbox_pad = metadata['letterbox_pad']
            img_original = metadata['img_original']

            detections = []
            max_conf = 0.0

            if len(det) > 0:

                # Prepare scaling parameters
                gn = torch.tensor(scaling_shape)[[1, 0, 1, 0]]

                if 'classic' in self.compatibility_mode:
                    ratio = None
                    ratio_pad = None
                else:
                    ratio = (img_original.shape[0]/scaling_shape[0],
                             img_original.shape[1]/scaling_shape[1])
                    ratio_pad = (ratio, letterbox_pad)

                # Rescale boxes
                if 'classic' in self.compatibility_mode:
                    det[:, :4] = scale_coords(batch_tensor.shape[2:], det[:, :4], img_original.shape).round()
                else:
                    det[:, :4] = scale_coords(batch_tensor.shape[2:], det[:, :4], scaling_shape, ratio_pad).round()

                # Process each detection
                for *xyxy, conf, cls in reversed(det):
                    if conf < detection_threshold:
                        continue

                    # Convert to YOLO format then to MD format
                    xywh = (xyxy2xywh(torch.tensor(xyxy).view(1, 4)) / gn).view(-1).tolist()
                    api_box = ct_utils.convert_yolo_to_xywh(xywh)

                    if 'classic' in self.compatibility_mode:
                        api_box = ct_utils.truncate_float_array(api_box, precision=COORD_DIGITS)
                        conf = ct_utils.truncate_float(conf.tolist(), precision=CONF_DIGITS)
                    else:
                        api_box = ct_utils.round_float_array(api_box, precision=COORD_DIGITS)
                        conf = ct_utils.round_float(conf.tolist(), precision=CONF_DIGITS)

                    if not self.use_model_native_classes:
                        cls = int(cls.tolist()) + 1
                        if cls not in (1, 2, 3):
                            raise KeyError(f'{cls} is not a valid class.')
                    else:
                        cls = int(cls.tolist())

                    detections.append({
                        'category': str(cls),
                        'conf': conf,
                        'bbox': api_box
                    })
                    max_conf = max(max_conf, conf)

                # ...for each detection

            # ...if there are > 0 detections

            # Store result for this image
            results[original_idx] = {
                'file': current_image_id,
                'detections': detections,
                'max_detection_conf': max_conf
            }

        # ...for each image

    # ...def _process_batch_group(...)

    def generate_detections_one_image(self,
                                      img_original,
                                      image_id='unknown',
                                      detection_threshold=0.00001,
                                      image_size=None,
                                      augment=False,
                                      verbose=False):
        """
        Run a detector on an image (wrapper around batch function).

        Args:
            img_original (Image, np.array, or dict): the image on which we should run the detector, with
                EXIF rotation already handled, or a dict representing a preprocessed image with associated
                letterbox parameters
            image_id (str, optional): a path to identify the image; will be in the "file" field
                of the output object
            detection_threshold (float, optional): only detections above this confidence threshold
                will be included in the return value
            image_size (int, optional): image size (long side) to use for inference, or None to
                use the default size specified at the time the model was loaded
            augment (bool, optional): enable (implementation-specific) image augmentation
            verbose (bool, optional): enable additional debug output

        Returns:
            dict: a dictionary with the following fields:
                - 'file' (filename, always present)
                - 'max_detection_conf' (removed from MegaDetector output files by default, but generated here)
                - 'detections' (a list of detection objects containing keys 'category', 'conf', and 'bbox')
                - 'failure' (a failure string, or None if everything went fine)
        """

        # Prepare batch inputs
        if isinstance(img_original, dict):
            batch_results = self.generate_detections_one_batch(
                img_original=[img_original],
                image_id=None,
                detection_threshold=detection_threshold,
                image_size=image_size,
                augment=augment,
                verbose=verbose)
        else:
            batch_results = self.generate_detections_one_batch(
                img_original=[img_original],
                image_id=[image_id],
                detection_threshold=detection_threshold,
                image_size=image_size,
                augment=augment,
                verbose=verbose)

        # Return the single result
        return batch_results[0]

    # ...def generate_detections_one_image(...)

# ...class PTDetector



"""

create_lila_blank_set.py

Create a folder of blank images sampled from LILA.  We'll aim for diversity, so less-common
locations will be oversampled relative to more common locations.  We'll also run MegaDetector
(with manual review) to remove some incorrectly-labeled, not-actually-empty images from our
blank set.

We'll store location information for each image in a .json file, so we can split locations
into train/val in downstream tasks.

"""

#%% Constants and imports

import os
import random
import math
import json

import numpy as np
from tqdm import tqdm
from multiprocessing.pool import ThreadPool
from urllib.parse import urlparse
from collections import defaultdict

from megadetector.data_management.lila.lila_common import read_lila_all_images_file
from megadetector.utils.url_utils import download_url
from megadetector.utils.ct_utils import sort_dictionary_by_value
from megadetector.utils.path_utils import is_image_file
from megadetector.utils.path_utils import find_images
from megadetector.visualization import visualization_utils as vis_utils
from megadetector.utils.path_utils import recursive_file_list
from megadetector.utils import ct_utils


#%% Environment

# We'll write images, metadata downloads, and temporary files here
lila_local_base = os.path.expanduser('~/lila')

metadata_dir = os.path.join(lila_local_base,'metadata')
os.makedirs(metadata_dir,exist_ok=True)

project_base = os.path.join(lila_local_base,'lila_blanks')

candidate_blanks_base = os.path.join(project_base,'candidate_blanks')
os.makedirs(candidate_blanks_base,exist_ok=True)

confirmed_blanks_base = os.path.join(project_base,'confirmed_blanks')
os.makedirs(confirmed_blanks_base,exist_ok=True)

md_possible_non_blanks_folder = os.path.join(project_base,'candidate_non_blanks')
os.makedirs(md_possible_non_blanks_folder,exist_ok=True)

location_to_blank_image_urls_cache_file = os.path.join(project_base,
                                                    'location_to_blank_image_urls.json')

md_results_file = os.path.join(project_base,'lila_blanks_md_results.json')

all_fn_relative_to_location_file = os.path.join(project_base,'all_fn_relative_to_location.json')
confirmed_fn_relative_to_location_file = os.path.join(project_base,'confirmed_fn_relative_to_location.json')

preferred_image_download_source = 'gcp'

# Number of concurrent download threads
n_download_threads = 20

n_blanks = 100000

random.seed(0)


#%% Download and open the giant table of image URLs and labels

# ~60 seconds to download, unzip, and open
df = read_lila_all_images_file(metadata_dir)


#%% Explore blank labels

# Original labels we're treating as blank:
blank_original_labels = (
    'empty','misfire'
)

# Some notable original labels we're *not* treating as blank:
nonblank_original_labels = (
    'unclassifiable', 'unidentifiable', 'unidentified', 'unknown', 'fire',
    'foggy lens', 'foggy weather', 'blurred', 'end', 'eye_shine', 'ignore',
    'lens obscured', 'misdirected', 'other', 'start', 'sun', 'problem',
    'tilted', 'vegetation obstruction', 'snow on lens', 'malfunction'
)

other_labels_without_common_names = (
    'car', 'motorcycle', 'vehicle'
)

common_names = sorted(list(df['common_name'].unique()),
                    key=lambda x:str(x) if isinstance(x,float) else x)
original_labels = sorted(list(df['original_label'].unique()),
                        key=lambda x:str(x) if isinstance(x,float) else x)

# Blanks are represented as NaN in the "common_name" column (though not all NaN's are blanks)
assert '' not in common_names
assert all([s not in common_names for s in blank_original_labels])
assert all([s not in common_names for s in nonblank_original_labels])
assert np.nan in common_names

# Blanks are represented as "empty" or "misfire" in the "original_label" column
assert all([s in original_labels for s in blank_original_labels])
assert all([s in original_labels for s in nonblank_original_labels])
assert all([s in original_labels for s in other_labels_without_common_names])
assert all([s not in original_labels for s in ('','blank','none',np.nan)])


#%% Count empty labels and common names

common_names_with_empty_original_labels = set()
original_labels_with_nan_common_names = set()

common_name_to_count = defaultdict(int)
original_label_to_count = defaultdict(int)

# This loop takes ~10 mins
for i_row,row in tqdm(df.iterrows(),total=len(df)):

    common_name = row['common_name']
    original_label = row['original_label']

    if isinstance(common_name,float):
        assert np.isnan(common_name)
        original_labels_with_nan_common_names.add(original_label)

    common_name = str(common_name)

    assert isinstance(original_label,str)
    if original_label in blank_original_labels:
        common_names_with_empty_original_labels.add(common_name)
    common_name_to_count[common_name] += 1
    original_label_to_count[original_label] += 1


#%% Look at the most common labels and common names

common_name_to_count = sort_dictionary_by_value(common_name_to_count,reverse=True)
original_label_to_count = sort_dictionary_by_value(original_label_to_count,reverse=True)

k = 10

print('\nMost frequent common names:\n')

i_label = 0
for i_label,s in enumerate(common_name_to_count):
    if i_label >= k:
        break
    print('{}: {}'.format(s,common_name_to_count[s]))

print('\nMost frequent original labels:\n')

i_label = 0
for i_label,s in enumerate(original_label_to_count):
    if i_label >= k:
        break
    print('{}: {}'.format(s,original_label_to_count[s]))


#%% Do some consistency checks over the empty labels and stats

# All images called 'empty' should have NaN as their common name
assert (len(common_names_with_empty_original_labels) == 1)
assert next(iter(common_names_with_empty_original_labels)) == 'nan'

# 'empty' should be the most frequent original label overall
assert next(iter(original_label_to_count)) == 'empty'

# NaN should be the most frequent common name overall
assert next(iter(common_name_to_count)) == 'nan'

for s in original_labels_with_nan_common_names:
    assert \
        (s in blank_original_labels) or \
        (s in nonblank_original_labels) or \
        (s in other_labels_without_common_names)


#%% Map locations to blank images

force_map_locations = False

# Load from .json if available
if (not force_map_locations) and (os.path.isfile(location_to_blank_image_urls_cache_file)):

    with open(location_to_blank_image_urls_cache_file,'r') as f:
        location_to_blank_image_urls = json.load(f)

else:

    location_to_blank_image_urls = defaultdict(list)

    # i_row = 0; row = df.iloc[i_row]
    for i_row,row in tqdm(df.iterrows(),total=len(df)):

        location_id = row['location_id']
        url = row['url']

        original_label = row['original_label']
        if original_label in blank_original_labels:
            assert np.isnan(row['common_name'])
            location_to_blank_image_urls[location_id].append(url)

    ct_utils.write_json(location_to_blank_image_urls_cache_file, location_to_blank_image_urls)

n_locations_with_blanks = len(location_to_blank_image_urls)
print('Found {} locations with blank images'.format(n_locations_with_blanks))


#%% Sample blanks

random.seed(0)

# Make a fresh copy of the lists
location_to_unsampled_blank_image_urls = {}

# location = next(iter(location_to_blank_image_urls.keys()))
for location in location_to_blank_image_urls:
    blank_image_urls_this_location = location_to_blank_image_urls[location]
    unsampled_blank_image_urls_this_location = blank_image_urls_this_location.copy()
    location_to_unsampled_blank_image_urls[location] = unsampled_blank_image_urls_this_location

# Put locations in a random order
location_ids = list(location_to_unsampled_blank_image_urls.keys())
random.shuffle(location_ids)

blank_urls = []
location_to_sampled_blanks = defaultdict(list)
fully_sampled_locations = set()

# Pick from each location until we hit our limit or have no blanks left
while(True):

    found_sample = False

    # location = location_ids[0]
    for location in location_ids:

        unsampled_images_this_location = location_to_unsampled_blank_image_urls[location]
        if len(unsampled_images_this_location) == 0:
            fully_sampled_locations.add(location)
            continue

        url = random.choice(unsampled_images_this_location)
        blank_urls.append(url)
        location_to_unsampled_blank_image_urls[location].remove(url)
        location_to_sampled_blanks[location].append(url)
        found_sample = True

        if len(blank_urls) == n_blanks:
            break

    # ...for each location

    if not found_sample:
        print('Terminating after {} blanks, we ran out before hitting {}'.format(
            len(blank_urls),n_blanks))

    if len(blank_urls) == n_blanks:
        break

# ...while(True)

assert len(blank_urls) <= n_blanks
min_blanks_per_location = math.floor(n_blanks/n_locations_with_blanks)
max_blanks_per_location = -1
for location in location_to_sampled_blanks:
    n_blanks_this_location = len(location_to_sampled_blanks[location])
    if n_blanks_this_location >= max_blanks_per_location:
        max_blanks_per_location = n_blanks_this_location
    assert (location in fully_sampled_locations) or \
        n_blanks_this_location >= min_blanks_per_location

print('Choose {} blanks from {} locations'.format(n_blanks,len(location_ids)))
print('Fully sampled {} locations'.format(len(fully_sampled_locations)))
print('Max samples per location: {}'.format(max_blanks_per_location))


#%% Download those image files (prep)

container_to_url_base = {
                        'lilawildlife.blob.core.windows.net':'/lila-wildlide/',
                        'storage.googleapis.com':'/public-datasets-lila/'
                        }

def download_relative_filename(url, output_base, verbose=False, url_base=None, overwrite=False):
    """
    Download a URL to output_base, preserving relative path
    """

    result = {'status':'unknown','url':url,'destination_filename':None}

    if url_base is None:
        assert url.startswith('https://')
        container = url.split('/')[2]
        assert container in container_to_url_base
        url_base = container_to_url_base[container]

    assert url_base.startswith('/') and url_base.endswith('/')

    p = urlparse(url)
    relative_filename = str(p.path)
    # remove the leading '/'
    assert relative_filename.startswith(url_base)
    relative_filename = relative_filename.replace(url_base,'',1)

    destination_filename = os.path.join(output_base,relative_filename)
    result['destination_filename'] = destination_filename

    if ((os.path.isfile(destination_filename)) and (not overwrite)):
        result['status'] = 'skipped'
        return result
    try:
        download_url(url, destination_filename, verbose=verbose)
    except Exception as e:
        print('Warning: error downloading URL {}: {}'.format(
            url,str(e)))
        result['status'] = 'error: {}'.format(str(e))
        return result

    result['status'] = 'success'
    return result


def azure_url_to_gcp_http_url(url,error_if_not_azure_url=True):
    """
    Most URLs point to Azure by default, but most files are available on both Azure and GCP.
    This function converts an Azure URL to the corresponding GCP http:// url.
    """

    lila_azure_storage_account = 'https://lilawildlife.blob.core.windows.net'
    gcp_bucket_api_url = 'https://storage.googleapis.com/public-datasets-lila'
    error_if_not_azure_url = False

    if error_if_not_azure_url:
        assert url.startswith(lila_azure_storage_account)
    gcp_url = url.replace(lila_azure_storage_account,gcp_bucket_api_url,1)
    return gcp_url

# Convert Azure URLs to GCP URLs if necessary
if preferred_image_download_source != 'azure':
    assert preferred_image_download_source == 'gcp'
    blank_urls = [azure_url_to_gcp_http_url(url) for url in blank_urls]


#%% Download those image files (execution)

print('Downloading {} images on {} workers'.format(len(blank_urls),n_download_threads))

if n_download_threads <= 1:

    results = []

    # url = all_urls[0]
    for url in tqdm(blank_urls):
        results.append(download_relative_filename(url,candidate_blanks_base,url_base=None))

else:

    pool = ThreadPool(n_download_threads)
    results = list(tqdm(pool.imap(lambda s: download_relative_filename(
        s,candidate_blanks_base,url_base=None),
        blank_urls), total=len(blank_urls)))

# pool.terminate()


#%% Review results

error_urls = []
for r in results:
    if r['status'] != 'success':
        error_urls.append(r['url'])

print('Errors on {} of {} downloads'.format(len(error_urls),len(results)))


#%% Run MegaDetector on the folder

cmd = 'python run_detector_batch.py MDV5A "{}" "{}"'.format(
    candidate_blanks_base,md_results_file)
cmd += ' --recursive --output_relative_filenames'

# import clipboard; clipboard.copy(cmd); print(cmd)


#%% Review MD results that suggests images are non-empty

assert os.path.isfile(md_results_file)

category_name_to_threshold = {'animal':0.25,'person':0.25,'vehicle':0.25}
min_threshold = min(category_name_to_threshold.values())
with open(md_results_file,'r') as f:
    md_results = json.load(f)

images_to_review_to_detections = {}

category_id_to_threshold = {}
for category_id in md_results['detection_categories']:
    category_name = md_results['detection_categories'][category_id]
    category_id_to_threshold[category_id] = category_name_to_threshold[category_name]

# im = md_results['images'][0]
for im in md_results['images']:

    if 'detections' not in im:
        continue

    found_object = False
    for det in im['detections']:
        threshold = category_id_to_threshold[det['category']]
        if det['conf'] >= threshold:
            found_object = True
            break
    if found_object:
        images_to_review_to_detections[im['file']] = im['detections']

print('Flagging {} of {} images for review'.format(len(images_to_review_to_detections),len(md_results['images'])))

output_file_to_source_file = {}

# i_fn = 0; source_file_relative = images_to_review[i_fn]
for i_fn,source_file_relative in tqdm(enumerate(images_to_review_to_detections),
                                    total=len(images_to_review_to_detections)):

    source_file_abs = os.path.join(candidate_blanks_base,source_file_relative)
    assert os.path.isfile(source_file_abs)
    ext = os.path.splitext(source_file_abs)[1]
    target_file_relative = str(i_fn).zfill(8) + ext
    target_file_abs = os.path.join(md_possible_non_blanks_folder,target_file_relative)
    output_file_to_source_file[target_file_relative] = source_file_relative
    # shutil.copyfile(source_file_abs,target_file_abs)
    vis_utils.draw_bounding_boxes_on_file(input_file=source_file_abs,
                                        output_file=target_file_abs,
                                        detections=images_to_review_to_detections[source_file_relative],
                                        confidence_threshold=min_threshold,
                                        target_size=(1280,-1))

# This is a temporary file I just used during debugging
ct_utils.write_json(os.path.join(project_base,'output_file_to_source_file.json'), output_file_to_source_file)


#%% Manual review

# Delete images that are *not* empty


#%% Figure out which images are still there; these are the actually-blank ones

remaining_images = set(os.listdir(md_possible_non_blanks_folder))
print('Kept {} of {} candidate blank images'.format(len(remaining_images),
                                                    len(images_to_review_to_detections)))

removed_blank_images_relative = []

# output_file = next(iter(output_file_to_source_file.keys()))
for output_file in tqdm(output_file_to_source_file.keys()):
    if output_file not in remaining_images:
        source_file_relative = output_file_to_source_file[output_file]
        removed_blank_images_relative.append(source_file_relative)

removed_blank_images_relative_set = set(removed_blank_images_relative)
assert len(removed_blank_images_relative) + len(remaining_images) == len(output_file_to_source_file)


#%% Copy only the confirmed blanks to the confirmed folder

all_candidate_blanks = recursive_file_list(candidate_blanks_base,return_relative_paths=True)
print('Found {} candidate blanks'.format(len(all_candidate_blanks)))

skipped_images_relative = []
skipped_non_images = []

for source_fn_relative in tqdm(all_candidate_blanks):

    # Skip anything we removed from the "candidate non-blanks" folder; these weren't really
    # blank.
    if source_fn_relative in removed_blank_images_relative_set:
        skipped_images_relative.append(source_fn_relative)
        continue

    if not is_image_file(source_fn_relative):
        # Not a typo; "skipped images" really means "skipped files"
        skipped_images_relative.append(source_fn_relative)
        skipped_non_images.append(source_fn_relative)


    source_fn_abs = os.path.join(candidate_blanks_base,source_fn_relative)
    assert os.path.isfile(source_fn_abs)
    target_fn_abs = os.path.join(confirmed_blanks_base,source_fn_relative)
    os.makedirs(os.path.dirname(target_fn_abs),exist_ok=True)
    # shutil.copyfile(source_fn_abs,target_fn_abs)

print('Skipped {} files ({} non-image files)'.format(len(skipped_images_relative),
                                                    len(skipped_non_images)))


#%% Validate the folder of confirmed blanks

# all_confirmed_blanks = recursive_file_list(confirmed_blanks_base,return_relative_paths=True)
all_confirmed_blanks = find_images(confirmed_blanks_base,return_relative_paths=True,recursive=True)
assert len(all_confirmed_blanks) < len(all_candidate_blanks)
print('Found {} confirmed blanks'.format(len(all_confirmed_blanks)))


#%% Manually review a few of the images we skipped

# ...to make sure they're non-blank
i_image = random.randint(0, len(skipped_images_relative))
fn_relative = skipped_images_relative[i_image]
fn_abs = os.path.join(candidate_blanks_base,fn_relative)
assert os.path.isfile(fn_abs)

# import clipboard; clipboard.copy('feh --scale-down "{}"'.format(fn_abs))


#%% Record location information for each confirmed file

# Map every URL's path to the corresponding location
#
# This is *all empty URLs*, not just the ones we downloaded
all_fn_relative_to_location = {}

# location = next(iter(location_to_blank_image_urls.keys()))
for location in tqdm(location_to_blank_image_urls):
    urls_this_location = location_to_blank_image_urls[location]

    # url = urls_this_location[0]
    for url in urls_this_location:
        # Turn:
        #
        # https://lilablobssc.blob.core.windows.net/caltech-unzipped/cct_images/5968c0f9-23d2-11e8-a6a3-ec086b02610b.jpg'
        #
        # ...into:
        #
        # caltech-unzipped/cct_images/5968c0f9-23d2-11e8-a6a3-ec086b02610b.jpg'
        p = urlparse(url)
        fn_relative = str(p.path)[1:]
        all_fn_relative_to_location[fn_relative] = location

# Build a much smaller mapping of just the confirmed blanks
confirmed_fn_relative_to_location = {}
for i_fn,fn_relative in tqdm(enumerate(all_confirmed_blanks),total=len(all_confirmed_blanks)):
    confirmed_fn_relative_to_location[fn_relative] = all_fn_relative_to_location[fn_relative]

ct_utils.write_json(all_fn_relative_to_location_file, all_fn_relative_to_location)
ct_utils.write_json(confirmed_fn_relative_to_location_file, confirmed_fn_relative_to_location)



"""

write_html_image_list.py

Given a list of image file names, writes an HTML file that
shows all those images, with optional one-line headers above each.

Each "filename" can also be a dict with elements 'filename','title',
'imageStyle','textStyle', 'linkTarget'

"""

#%% Constants and imports

import os
import math
import urllib

from megadetector.utils import path_utils


#%% write_html_image_list

def write_html_image_list(filename=None,images=None,options=None):
    """
    Given a list of image file names, writes an HTML file that shows all those images,
    with optional one-line headers above each.

    Args:
        filename (str, optional): the .html output file; if None, just returns a valid
            options dict
        images (list, optional): the images to write to the .html file; if None, just returns
            a valid options dict.  This can be a flat list of image filenames, or this can
            be a list of dictionaries with one or more of the following fields:

            - filename (image filename) (required, all other fields are optional)
            - imageStyle (css style for this image)
            - textStyle (css style for the title associated with this image)
            - title (text label for this image)
            - linkTarget (URL to which this image should link on click)

        options (dict, optional): a dict with one or more of the following fields:

            - f_html (file pointer to write to, used for splitting write operations over multiple calls)
            - pageTitle (HTML page title)
            - headerHtml (html text to include before the image list)
            - subPageHeaderHtml (html text to include before the images when images are broken into pages)
            - trailerHtml (html text to include after the image list)
            - defaultImageStyle (default css style for images)
            - defaultTextStyle (default css style for image titles)
            - maxFiguresPerHtmlFile (max figures for a single HTML file; overflow will be handled by creating
              multiple files and a TOC with links)
            - urlEncodeFilenames (default True, e.g. '#' will be replaced by '%23')
            - urlEncodeLinkTargets (default True, e.g. '#' will be replaced by '%23')

    """

    # returns an options struct
    if options is None:
        options = {}

    if 'f_html' not in options:
        options['f_html'] = -1

    if 'pageTitle' not in options or options['pageTitle'] is None:
        options['pageTitle'] = ''

    if 'headerHtml' not in options or options['headerHtml'] is None:
        options['headerHtml'] = ''

    if 'subPageHeaderHtml' not in options or options['subPageHeaderHtml'] is None:
        options['subPageHeaderHtml'] = ''

    if 'trailerHtml' not in options or options['trailerHtml'] is None:
        options['trailerHtml'] = ''

    if 'defaultTextStyle' not in options or options['defaultTextStyle'] is None:
        options['defaultTextStyle'] = \
        "font-family:calibri,verdana,arial;font-weight:bold;font-size:150%;text-align:left;margin:0px;"

    if 'defaultImageStyle' not in options or options['defaultImageStyle'] is None:
        options['defaultImageStyle'] = \
        "margin:0px;margin-top:5px;margin-bottom:5px;"

    if 'urlEncodeFilenames' not in options or options['urlEncodeFilenames'] is None:
        options['urlEncodeFilenames'] = True

    if 'urlEncodeLinkTargets' not in options or options['urlEncodeLinkTargets'] is None:
        options['urlEncodeLinkTargets'] = True

    # Possibly split the html output for figures into multiple files; Chrome gets sad with
    # thousands of images in a single tab.
    if 'maxFiguresPerHtmlFile' not in options or options['maxFiguresPerHtmlFile'] is None:
        options['maxFiguresPerHtmlFile'] = math.inf

    if filename is None or images is None:
        return options

    # images may be a list of images or a list of image/style/title dictionaries,
    # enforce that it's the latter to simplify downstream code
    for i_image,image_info in enumerate(images):
        if isinstance(image_info,str):
            image_info = {'filename':image_info}
        if 'filename' not in image_info:
            image_info['filename'] = ''
        if 'imageStyle' not in image_info:
            image_info['imageStyle'] = options['defaultImageStyle']
        if 'title' not in image_info:
            image_info['title'] = ''
        if 'linkTarget' not in image_info:
            image_info['linkTarget'] = ''
        if 'textStyle' not in image_info:
            text_style = options['defaultTextStyle']
            image_info['textStyle'] = options['defaultTextStyle']
        images[i_image] = image_info

    n_images = len(images)

    # If we need to break this up into multiple files...
    if n_images > options['maxFiguresPerHtmlFile']:

        # You can't supply your own file handle in this case
        if options['f_html'] != -1:
            raise ValueError(
                    "You can't supply your own file handle if we have to page the image set")

        figure_file_starting_indices = list(range(0,n_images,options['maxFiguresPerHtmlFile']))

        assert len(figure_file_starting_indices) > 1

        # Open the meta-output file
        f_meta = open(filename,'w')

        # Write header stuff
        title_string = '<title>Index page</title>'
        if len(options['pageTitle']) > 0:
            title_string = '<title>Index page for: {}</title>'.format(options['pageTitle'])
        f_meta.write('<html><head>{}</head><body>\n'.format(title_string))
        f_meta.write(options['headerHtml'])
        f_meta.write('<table border = 0 cellpadding = 2>\n')

        for starting_index in figure_file_starting_indices:

            i_start = starting_index
            i_end = starting_index + options['maxFiguresPerHtmlFile'] - 1
            if i_end >= n_images:
                i_end = n_images-1

            trailer = 'image_{:05d}_{:05d}'.format(i_start,i_end)
            local_figures_html_filename = path_utils.insert_before_extension(filename,trailer)
            f_meta.write('<tr><td>\n')
            f_meta.write('<p style="padding-bottom:0px;margin-bottom:0px;text-align:left;font-family:''segoe ui'',calibri,arial;font-size:100%;text-decoration:none;font-weight:bold;">') # noqa
            f_meta.write('<a href="{}">Figures for images {} through {}</a></p></td></tr>\n'.format(
                os.path.basename(local_figures_html_filename),i_start,i_end))

            local_images = images[i_start:i_end+1]

            local_options = options.copy()
            local_options['headerHtml'] = options['subPageHeaderHtml']
            local_options['trailerHtml'] = ''

            # Make a recursive call for this image set
            write_html_image_list(local_figures_html_filename,local_images,local_options)

        # ...for each page of images

        f_meta.write('</table></body>\n')
        f_meta.write(options['trailerHtml'])
        f_meta.write('</html>\n')
        f_meta.close()

        return options

    # ...if we have to make multiple sub-pages

    b_clean_up_file = False

    if options['f_html'] == -1:
        b_clean_up_file = True
        f_html = open(filename,'w')
    else:
        f_html = options['f_html']

    title_string = ''
    if len(options['pageTitle']) > 0:
        title_string = '<title>{}</title>'.format(options['pageTitle'])

    f_html.write('<html>{}<body>\n'.format(title_string))

    f_html.write(options['headerHtml'])

    # Write out images
    for i_image,image in enumerate(images):

        title = image['title']
        image_style = image['imageStyle']
        text_style = image['textStyle']
        filename = image['filename']
        link_target = image['linkTarget']

        # Remove unicode characters
        title = title.encode('ascii','ignore').decode('ascii')
        filename = filename.encode('ascii','ignore').decode('ascii')

        filename = filename.replace('\\','/')
        if options['urlEncodeFilenames']:
            filename = urllib.parse.quote(filename)

        if len(title) > 0:
            f_html.write(
                    '<p style="{}">{}</p>\n'\
                    .format(text_style,title))

        link_target = link_target.replace('\\','/')
        if options['urlEncodeLinkTargets']:
            # These are typically absolute paths, so we only want to mess with certain characters
            link_target = urllib.parse.quote(link_target,safe=':/')

        if len(link_target) > 0:
            f_html.write('<a href="{}">'.format(link_target))
            # image_style.append(';border:0px;')

        f_html.write('<img src="{}" style="{}">\n'.format(filename,image_style))

        if len(link_target) > 0:
            f_html.write('</a>')

        if i_image != len(images)-1:
            f_html.write('<br/>')

    # ...for each image we need to write

    f_html.write(options['trailerHtml'])

    f_html.write('</body></html>\n')

    if b_clean_up_file:
        f_html.close()

# ...function



"""

speciesnet_to_md.py

Converts the WI (SpeciesNet) predictions.json format to MD .json format.  This is just a
command-line wrapper around utils.wi_taxonomy_utils.generate_md_results_from_predictions_json.

"""

#%% Imports and constants

import sys
import argparse
from megadetector.utils.wi_taxonomy_utils import generate_md_results_from_predictions_json


#%% Command-line driver

def main(): # noqa

    parser = argparse.ArgumentParser()
    parser.add_argument('predictions_json_file', action='store', type=str,
                        help='.json file to convert from SpeciesNet predictions.json format to MD format')
    parser.add_argument('md_results_file', action='store', type=str,
                        help='output file to write in MD format')
    parser.add_argument('--base_folder', action='store', type=str, default=None,
                        help='leading string to remove from each path in the predictions.json ' + \
                            'file (to convert from absolute to relative paths)')

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    generate_md_results_from_predictions_json(args.predictions_json_file,
                                              args.md_results_file,
                                              args.base_folder)

if __name__ == '__main__':
    main()



"""

evaluate_model.py

Evaluate a species classifier.

Currently the implementation of multi-label multi-class classification is
non-functional.

Outputs the following files:

1) outputs_{split}.csv, one file per split, contains columns:
    - 'path': str, path to cropped image
    - 'label': str
    - 'weight': float
    - [label names]: float, confidence in each label

2) overall_metrics.csv, contains columns:
    - 'split': str
    - 'loss': float, mean per-example loss over entire epoch
    - 'acc_top{k}': float, accuracy@k over the entire epoch
    - 'loss_weighted' and 'acc_weighted_top{k}': float, weighted versions

3) confusion_matrices.npz
    - keys ['train', 'val', 'test']
    - values are np.ndarray, confusion matrices

4) label_stats.csv, per-label statistics, columns
    - 'split': str
    - 'label': str
    - 'precision': float
    - 'recall': float

"""

#%% Imports and constants

from __future__ import annotations

import argparse
from collections.abc import Mapping, Sequence
import json
import os
from pprint import pprint
from typing import Any, Optional

import numpy as np
import pandas as pd
import sklearn.metrics
import torch
import torchvision
import tqdm

from megadetector.classification import efficientnet, train_classifier
from megadetector.utils import ct_utils


#%% Example usage

"""
    python evaluate_model.py \
        $BASE_LOGDIR/$LOGDIR/params.json \
        $BASE_LOGDIR/$LOGDIR/ckpt_XX.pt \
        --output-dir $BASE_LOGDIR/$LOGDIR \
        --splits train val test \
        --batch-size 256
"""

SPLITS = ['train', 'val', 'test']


#%% Support functions

def check_override(params: Mapping[str, Any], key: str,
                   override: Optional[Any]) -> Any:
    """
    Return desired value, with optional override.
    """

    if override is None:
        return params[key]
    saved = params.get(key, None)
    print(f'Overriding saved {key}. Saved: {saved}. Override with: {override}.')
    return override


def trace_model(model_name: str, ckpt_path: str, num_classes: int,
                img_size: int) -> str:
    """
    Use TorchScript tracing to compile trained model into standalone file.

    For now, we have to use tracing instead of scripting. See
        https://github.com/lukemelas/EfficientNet-PyTorch/issues/89
        https://github.com/lukemelas/EfficientNet-PyTorch/issues/218

    Args:
        model_name: str
        ckpt_path: str, path to checkpoint file
        num_labels: int, number of classification classes
        img_size: int, size of input image, used for tracing

    Returns: str, name of file for compiled model. For example, if ckpt_path is
        '/path/to/ckpt_16.pt', then the returned path is
        '/path/to/ckpt_16_compiled.pt'.
    """

    root, ext = os.path.splitext(ckpt_path)
    compiled_path = root + '_compiled' + ext
    if os.path.exists(compiled_path):
        return compiled_path

    model = train_classifier.build_model(model_name, num_classes=num_classes,
                                         pretrained=ckpt_path, finetune=False)
    if 'efficientnet' in model_name:
        model.set_swish(memory_efficient=False)
    model.eval()

    ex_img = torch.rand(1, 3, img_size, img_size)
    scripted_model = torch.jit.trace(model, (ex_img,))

    scripted_model.save(compiled_path)
    print('Saved TorchScript compiled model to', compiled_path)
    return compiled_path


def calc_per_label_stats(cm: np.ndarray, label_names: Sequence[str]
                         ) -> pd.DataFrame:
    """
    Args:
        cm: np.ndarray, type int, confusion matrix C such that C[i,j] is the #
            of observations from group i that are predicted to be in group j
        label_names: list of str, label names in order of label id

    Returns: pd.DataFrame, index 'label', columns ['precision', 'recall']
        precision values are in [0, 1]
        recall values are in [0, 1], or np.nan if that label had 0 ground-truth
            observations
    """

    tp = np.diag(cm)  # true positives

    predicted_positives = cm.sum(axis=0, dtype=np.float64)  # tp + fp
    predicted_positives[predicted_positives == 0] += 1e-8

    all_positives = cm.sum(axis=1, dtype=np.float64)  # tp + fn
    all_positives[all_positives == 0] = np.nan

    df = pd.DataFrame()
    df['label'] = label_names
    df['precision'] = tp / predicted_positives
    df['recall'] = tp / all_positives
    df.set_index('label', inplace=True)
    return df


def test_epoch(model: torch.nn.Module,
               loader: torch.utils.data.DataLoader,
               weighted: bool,
               device: torch.device,
               label_names: Sequence[str],
               top: Sequence[int] = (1, 3),
               loss_fn: Optional[torch.nn.Module] = None,
               target_mapping: Mapping[int, Sequence[int]] = None
               ) -> tuple[pd.DataFrame, pd.Series, np.ndarray]:
    """
    Runs for 1 epoch.

    Args:
        model: torch.nn.Module
        loader: torch.utils.data.DataLoader
        weighted: bool, whether to calculate weighted accuracy statistics
        device: torch.device
        label_names: list of str, label names in order of label id
        top: tuple of int, list of values of k for calculating top-K accuracy
        loss_fn: optional loss function, calculates per-example loss
        target_mapping: optional dict, label_id => list of ids from classifier
            that should map to the label_id

    Returns:
        df: pd.DataFrame, columns ['img_file', 'label', 'weight', label_names]
        metrics: pd.Series, type float, index includes:
            'loss': mean per-example loss over entire epoch,
                only included if loss_fn is not None
            'acc_top{k}': accuracy@k over the entire epoch
            'loss_weighted' and 'acc_weighted_top{k}': weighted versions, only
                included if weighted=True
        cm: np.ndarray, confusion matrix C such that C[i,j] is the # of
            observations known to be in group i and predicted to be in group j
    """

    # set dropout and BN layers to eval mode
    model.eval()

    if loss_fn is not None:
        losses = train_classifier.AverageMeter()
    accuracies_topk = {k: train_classifier.AverageMeter() for k in top}  # acc@k
    if weighted:
        accs_weighted = {k: train_classifier.AverageMeter() for k in top}
        losses_weighted = train_classifier.AverageMeter()

    num_examples = len(loader.dataset)
    num_labels = len(label_names)

    all_img_files = []
    all_probs = np.zeros([num_examples, len(label_names)], dtype=np.float32)
    all_labels = np.zeros(num_examples, dtype=np.int32)
    if weighted:
        all_weights = np.zeros(num_examples, dtype=np.float32)

    batch_slice = slice(0, 0)
    tqdm_loader = tqdm.tqdm(loader)
    with torch.no_grad():
        for batch in tqdm_loader:
            if weighted:
                inputs, labels, img_files, weights = batch
            else:
                # even if batch contains sample weights, don't use them
                inputs, labels, img_files = batch[0:3]
                weights = None

            all_img_files.append(img_files)

            batch_size = labels.size(0)
            batch_slice = slice(batch_slice.stop, batch_slice.stop + batch_size)
            all_labels[batch_slice] = labels
            if weighted:
                all_weights[batch_slice] = weights
                weights = weights.to(device, non_blocking=True)

            inputs = inputs.to(device, non_blocking=True)
            labels = labels.to(device, non_blocking=True)
            outputs = model(inputs)

            # Do target mapping on the outputs (unnormalized logits) instead of
            # the normalized (softmax) probabilities, because the loss function
            # uses unnormalized logits. Summing probabilities is equivalent to
            # log-sum-exp of unnormalized logits.
            if target_mapping is not None:
                outputs_mapped = torch.zeros(
                    [batch_size, num_labels], dtype=outputs.dtype,
                    device=outputs.device)
                for target, cols in target_mapping.items():
                    outputs_mapped[:, target] = torch.logsumexp(
                        outputs[:, cols], dim=1)
                outputs = outputs_mapped

            probs = torch.nn.functional.softmax(outputs, dim=1).cpu()
            all_probs[batch_slice] = probs

            desc = []
            if loss_fn is not None:
                loss = loss_fn(outputs, labels)
                losses.update(loss.mean().item(), n=batch_size)
                desc.append(f'Loss {losses.val:.3f} ({losses.avg:.3f})')
                if weights is not None:
                    loss_weighted = (loss * weights).mean()
                    losses_weighted.update(loss_weighted.item(), n=batch_size)

            top_correct = train_classifier.correct(
                outputs, labels, weights=None, top=top)
            for k, acc in accuracies_topk.items():
                acc.update(top_correct[k] * (100. / batch_size), n=batch_size)
                desc.append(f'Acc@{k} {acc.val:.2f} ({acc.avg:.2f})')

            if weighted:
                top_correct = train_classifier.correct(
                    outputs, labels, weights=weights, top=top)
                for k, acc in accs_weighted.items():
                    acc.update(top_correct[k] * (100. / batch_size),
                               n=batch_size)
                    desc.append(f'Acc_w@{k} {acc.val:.2f} ({acc.avg:.2f})')

            tqdm_loader.set_description(' '.join(desc))

    # a confusion matrix C is such that C[i,j] is the # of observations known to
    # be in group i and predicted to be in group j.
    all_preds = all_probs.argmax(axis=1)
    cm = sklearn.metrics.confusion_matrix(
        y_true=all_labels, y_pred=all_preds, labels=np.arange(num_labels))

    df = pd.DataFrame()
    df['path'] = np.concatenate(all_img_files)
    df['label'] = list(map(label_names.__getitem__, all_labels))
    df['weight'] = all_weights
    df[label_names] = all_probs

    metrics = {}
    if loss_fn is not None:
        metrics['loss'] = losses.avg
        if weighted:
            metrics['loss_weighted'] = losses_weighted.avg
    for k, acc in accuracies_topk.items():
        metrics[f'acc_top{k}'] = acc.avg
    if weighted:
        for k, acc in accs_weighted.items():
            metrics[f'acc_weighted_top{k}'] = acc.avg
    return df, pd.Series(metrics), cm


#%% Main function

def main(params_json_path: str, ckpt_path: str, output_dir: str,
         splits: Sequence[str], target_mapping_json_path: Optional[str] = None,
         label_index_json_path: Optional[str] = None,
         **kwargs: Any) -> None:

    # input validation
    assert os.path.exists(params_json_path)
    assert os.path.exists(ckpt_path)
    assert (target_mapping_json_path is None) == (label_index_json_path is None)
    if target_mapping_json_path is not None:
        assert label_index_json_path is not None
        assert os.path.exists(target_mapping_json_path)
        assert os.path.exists(label_index_json_path)

    # Evaluating with accimage is much faster than Pillow or Pillow-SIMD, but accimage
    # is Linux-only.
    try:
        import accimage # noqa
        torchvision.set_image_backend('accimage')
    except:
        print('Warning: could not start accimage backend (ignore this if you\'re not using Linux)')

    # create output directory
    if not os.path.exists(output_dir):
        print('Creating output directory:', output_dir)
        os.makedirs(output_dir, exist_ok=True)

    with open(params_json_path, 'r') as f:
        params = json.load(f)
    pprint(params)

    # override saved params with kwargs
    for key, new in kwargs.items():
        if new is None:
            continue
        if key in params:
            saved = params[key]
            print(f'Overriding saved {key}. Saved: {saved}. '
                  f'Override with: {new}.')
        else:
            print(f'Did not find {key} in saved params. Using value {new}.')
        params[key] = new

    model_name: str = params['model_name']
    dataset_dir: str = params['dataset_dir']

    if 'efficientnet' in model_name:
        img_size = efficientnet.EfficientNet.get_image_size(model_name)
    else:
        img_size = 224

    # For now, we don't weight crops by detection confidence during
    # evaluation. But consider changing this.
    print('Creating dataloaders')
    loaders, label_names = train_classifier.create_dataloaders(
        dataset_csv_path=os.path.join(dataset_dir, 'classification_ds.csv'),
        label_index_json_path=os.path.join(dataset_dir, 'label_index.json'),
        splits_json_path=os.path.join(dataset_dir, 'splits.json'),
        cropped_images_dir=params['cropped_images_dir'],
        img_size=img_size,
        multilabel=params['multilabel'],
        label_weighted=params['label_weighted'],
        weight_by_detection_conf=False,
        batch_size=params['batch_size'],
        num_workers=params['num_workers'],
        augment_train=False)
    num_labels = len(label_names)

    # create model, compile with TorchScript if given checkpoint is not compiled
    print('Loading model from checkpoint')
    try:
        model = torch.jit.load(ckpt_path, map_location='cpu')
    except RuntimeError:
        compiled_path = trace_model(model_name, ckpt_path, num_labels, img_size)
        model = torch.jit.load(compiled_path, map_location='cpu')
    model, device = train_classifier.prep_device(model)

    if len(splits) == 0:
        print('No splits given! Exiting.')
        return

    target_cols_map = None
    if target_mapping_json_path is not None:
        assert label_index_json_path is not None

        # verify that target names matches original "label names" from dataset
        with open(target_mapping_json_path, 'r') as f:
            target_names_map = json.load(f)
        target_names = set(target_names_map.keys())

        # if the dataset does not already have a 'other' category, then the
        # 'other' category must come last in label_names to avoid conflicting
        # with an existing label_id
        if target_names != set(label_names):
            assert target_names == set(label_names) | {'other'}
            label_names.append('other')

            ct_utils.write_json(os.path.join(output_dir, 'label_index.json'), dict(enumerate(label_names)), indent=None)

        with open(label_index_json_path, 'r') as f:
            idx_to_label = json.load(f)
        classifier_name_to_idx = {
            idx_to_label[str(k)]: k for k in range(len(idx_to_label))
        }

        target_cols_map = {}
        for i_target, label_name in enumerate(label_names):
            classifier_names = target_names_map[label_name]
            target_cols_map[i_target] = [
                classifier_name_to_idx[classifier_name]
                for classifier_name in classifier_names
            ]

    # define loss function (criterion)
    loss_fn: torch.nn.Module
    if params['multilabel']:
        loss_fn = torch.nn.BCEWithLogitsLoss(reduction='none').to(device)
    else:
        loss_fn = torch.nn.CrossEntropyLoss(reduction='none').to(device)

    split_metrics = {}
    split_label_stats = {}
    cms = {}
    for split in splits:
        print(f'Evaluating {split}...')
        df, metrics, cm = test_epoch(
            model, loaders[split], weighted=True, device=device,
            label_names=label_names, loss_fn=loss_fn,
            target_mapping=target_cols_map)

        # this file ends up being huge, so we GZIP compress it
        output_csv_path = os.path.join(output_dir, f'outputs_{split}.csv.gz')
        df.to_csv(output_csv_path, index=False, compression='gzip')

        split_metrics[split] = metrics
        cms[split] = cm
        split_label_stats[split] = calc_per_label_stats(cm, label_names)

        # double check that the accuracy metrics are computed properly
        preds = df[label_names].to_numpy().argmax(axis=1)
        preds = np.asarray(label_names)[preds]
        assert np.isclose(metrics['acc_top1'] / 100.,
                          sum(preds == df['label']) / len(df))
        assert np.isclose(metrics['acc_weighted_top1'] / 100.,
                          sum((preds == df['label']) * df['weight']) / len(df))

    metrics_df = pd.concat(split_metrics, names=['split']).unstack(level=1)
    metrics_df.to_csv(os.path.join(output_dir, 'overall_metrics.csv'))

    # save the confusion matrices to .npz
    npz_path = os.path.join(output_dir, 'confusion_matrices.npz')
    np.savez_compressed(npz_path, **cms)

    # save per-label statistics
    label_stats_df = pd.concat(
        split_label_stats, names=['split', 'label']).reset_index()
    label_stats_csv_path = os.path.join(output_dir, 'label_stats.csv')
    label_stats_df.to_csv(label_stats_csv_path, index=False)


#%% Command-line driver

def _parse_args() -> argparse.Namespace:

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Evaluate trained model.')
    parser.add_argument(
        'params_json',
        help='path to params.json')
    parser.add_argument(
        'ckpt_path',
        help='path to checkpoint file (normal or TorchScript-compiled)')
    parser.add_argument(
        '-o', '--output-dir', required=True,
        help='(required) path to output directory')
    parser.add_argument(
        '--splits', nargs='*', choices=SPLITS, default=[],
        help='which splits to evaluate model on. If no splits are given, then '
             'only compiles normal checkpoint file using TorchScript without '
             'actually running the model.')

    other_classifier = parser.add_argument_group(
        'arguments for evaluating a model (e.g., MegaClassifier) on a '
        'different set of labels than what the model was trained on')
    other_classifier.add_argument(
        '--target-mapping',
        help='path to JSON file mapping target categories to classifier labels')
    other_classifier.add_argument(
        '--label-index',
        help='path to label index JSON file for classifier')

    override_group = parser.add_argument_group(
        'optional arguments to override values in params_json')
    override_group.add_argument(
        '--model-name',
        help='which EfficientNet or Resnet model')
    override_group.add_argument(
        '--batch-size', type=int,
        help='batch size for evaluating model')
    override_group.add_argument(
        '--num-workers', type=int,
        help='number of workers for data loading')
    override_group.add_argument(
        '--dataset-dir',
        help='path to directory containing classification_ds.csv, '
             'label_index.json, and splits.json')
    return parser.parse_args()


if __name__ == '__main__':

    args = _parse_args()
    main(params_json_path=args.params_json, ckpt_path=args.ckpt_path,
         output_dir=args.output_dir, splits=args.splits,
         target_mapping_json_path=args.target_mapping,
         label_index_json_path=args.label_index,
         model_name=args.model_name, batch_size=args.batch_size,
         num_workers=args.num_workers, dataset_dir=args.dataset_dir)



"""

generate_crops_from_cct.py

Given a .json file in COCO Camera Traps format, creates a cropped image for
each bounding box.

"""

#%% Imports and constants

import os
import argparse
import json

from tqdm import tqdm
from PIL import Image


#%% Functions

def generate_crops_from_cct(cct_file,image_dir,output_dir,padding=0,flat_output=True):
    """
    Given a .json file in COCO Camera Traps format, creates a cropped image for
    each bounding box.

    Args:
        cct_file (str): the COCO .json file from which we should load data
        image_dir (str): the folder where the images live; filenames in the .json
            file should be relative to this folder
        output_dir (str): the folder where we should write cropped images
        padding (float, optional): number of pixels we should expand each box before
            cropping
        flat_output (bool, optional): if False, folder structure will be preserved
            in the output, e.g. the image a/b/c/d.jpg will result in image files
            in the output folder called, e.g., a/b/c/d_crop_000_id_12345.jpg.  If
            [flat_output] is True, the corresponding output image will be
            a_b_c_d_crop_000_id_12345.jpg.
    """

    ## Read and validate input

    assert os.path.isfile(cct_file)
    assert os.path.isdir(image_dir)
    os.makedirs(output_dir,exist_ok=True)

    with open(cct_file,'r') as f:
        d = json.load(f)


    ## Find annotations for each image

    from collections import defaultdict

    # This actually maps image IDs to annotations, but only to annotations
    # containing boxes
    image_id_to_boxes = defaultdict(list)

    n_boxes = 0

    for ann in d['annotations']:
        if 'bbox' in ann:
            image_id_to_boxes[ann['image_id']].append(ann)
            n_boxes += 1

    print('Found {} boxes in {} annotations for {} images'.format(
        n_boxes,len(d['annotations']),len(d['images'])))


    ## Generate crops

    # im = d['images'][0]
    for im in tqdm(d['images']):

        input_image_fn = os.path.join(os.path.join(image_dir,im['file_name']))
        assert os.path.isfile(input_image_fn), 'Could not find image {}'.format(input_image_fn)

        if im['id'] not in image_id_to_boxes:
            continue

        annotations_this_image = image_id_to_boxes[im['id']]

        # Load the image
        img = Image.open(input_image_fn)

        # Generate crops
        # i_ann = 0; ann = annotations_this_image[i_ann]
        for i_ann,ann in enumerate(annotations_this_image):

            # x/y/w/h, origin at the upper-left
            bbox = ann['bbox']

            xmin = bbox[0]
            ymin = bbox[1]
            xmax = xmin + bbox[2]
            ymax = ymin + bbox[3]

            xmin -= padding / 2
            ymin -= padding / 2
            xmax += padding / 2
            ymax += padding / 2

            xmin = max(xmin,0)
            ymin = max(ymin,0)
            xmax = min(xmax,img.width-1)
            ymax = min(ymax,img.height-1)

            crop = img.crop(box=[xmin, ymin, xmax, ymax])

            output_fn = os.path.splitext(im['file_name'])[0].replace('\\','/')
            if flat_output:
                output_fn = output_fn.replace('/','_')
            output_fn = output_fn + '_crop' + str(i_ann).zfill(3) + '_id_' + ann['id']
            output_fn = output_fn + '.jpg'

            output_full_path = os.path.join(output_dir,output_fn)

            if not flat_output:
                os.makedirs(os.path.dirname(output_full_path),exist_ok=True)

            crop.save(output_full_path)

        # ...for each box

    # ...for each image

# ...generate_crops_from_cct()


#%% Interactive driver

if False:

    pass

    #%%

    cct_file = os.path.expanduser('~/data/noaa/noaa_estuary_fish.json')
    image_dir = os.path.expanduser('~/data/noaa/JPEGImages')
    padding = 50
    flat_output = True
    output_dir = '/home/user/tmp/noaa-fish-crops'

    generate_crops_from_cct(cct_file,image_dir,output_dir,padding,flat_output=True)
    files = os.listdir(output_dir)


#%% Command-line driver

def main():
    """
    Command-line interface to generate crops from a COCO Camera Traps .json file.
    """

    parser = argparse.ArgumentParser(
        description='Generate cropped images from a COCO Camera Traps .json file'
    )
    parser.add_argument(
        'cct_file',
        type=str,
        help='COCO .json file to load data from'
    )
    parser.add_argument(
        'image_dir',
        type=str,
        help='Folder where images are located'
    )
    parser.add_argument(
        'output_dir',
        type=str,
        help='Folder to which we should write cropped images'
    )
    parser.add_argument(
        '--padding',
        type=int,
        default=0,
        help='Pixels to expand each box before cropping'
    )
    parser.add_argument(
        '--flat_output',
        action='store_true',
        help='Flatten folder structure in output (preserves folder structure by default)'
    )

    args = parser.parse_args()

    generate_crops_from_cct(
        cct_file=args.cct_file,
        image_dir=args.image_dir,
        output_dir=args.output_dir,
        padding=args.padding,
        flat_output=args.flat_output
    )

    print(f'Generated crops in {args.output_dir}')

if __name__ == '__main__':
    main()
